#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest

from airflow.exceptions import TaskDeferred
from airflow.providers.amazon.aws.hooks.sagemaker import SageMakerHook
from airflow.providers.amazon.aws.operators.sagemaker import (
    SageMakerStartPipelineOperator,
    SageMakerStopPipelineOperator,
)
from airflow.providers.amazon.aws.triggers.sagemaker import SageMakerPipelineTrigger

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if TYPE_CHECKING:
    from unittest.mock import MagicMock


class TestSageMakerStartPipelineOperator:
    @mock.patch.object(SageMakerHook, "start_pipeline")
    @mock.patch.object(SageMakerHook, "check_status")
    def test_execute(self, check_status, start_pipeline):
        op = SageMakerStartPipelineOperator(
            task_id="test_sagemaker_operator",
            pipeline_name="my_pipeline",
            display_name="test_disp_name",
            pipeline_params={"is_a_test": "yes"},
            wait_for_completion=True,
            check_interval=12,
            verbose=False,
        )

        op.execute({})

        start_pipeline.assert_called_once_with(
            pipeline_name="my_pipeline",
            display_name="test_disp_name",
            pipeline_params={"is_a_test": "yes"},
        )
        check_status.assert_called_once()

    @mock.patch.object(SageMakerHook, "start_pipeline")
    def test_defer(self, start_mock):
        op = SageMakerStartPipelineOperator(
            task_id="test_sagemaker_operator",
            pipeline_name="my_pipeline",
            deferrable=True,
        )

        with pytest.raises(TaskDeferred) as defer:
            op.execute({})

        assert isinstance(defer.value.trigger, SageMakerPipelineTrigger)
        assert defer.value.trigger.waiter_type == SageMakerPipelineTrigger.Type.COMPLETE

    def test_template_fields(self):
        operator = SageMakerStartPipelineOperator(
            task_id="test_sagemaker_operator",
            pipeline_name="my_pipeline",
        )
        validate_template_fields(operator)


class TestSageMakerStopPipelineOperator:
    @mock.patch.object(SageMakerHook, "stop_pipeline")
    def test_execute(self, stop_pipeline):
        op = SageMakerStopPipelineOperator(
            task_id="test_sagemaker_operator", pipeline_exec_arn="pipeline_arn"
        )

        op.execute({})

        stop_pipeline.assert_called_once_with(
            pipeline_exec_arn="pipeline_arn",
            fail_if_not_running=False,
        )

    @mock.patch.object(SageMakerHook, "stop_pipeline")
    def test_defer(self, stop_mock: MagicMock):
        stop_mock.return_value = "Stopping"
        op = SageMakerStopPipelineOperator(
            task_id="test_sagemaker_operator",
            pipeline_exec_arn="my_pipeline_arn",
            deferrable=True,
        )

        with pytest.raises(TaskDeferred) as defer:
            op.execute({})

        assert isinstance(defer.value.trigger, SageMakerPipelineTrigger)
        assert defer.value.trigger.waiter_type == SageMakerPipelineTrigger.Type.STOPPED

    def test_template_fields(self):
        operator = SageMakerStopPipelineOperator(
            task_id="test_sagemaker_operator",
            pipeline_exec_arn="my_pipeline_arn",
            deferrable=True,
        )
        validate_template_fields(operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.dynamodb import DynamoDBHook
from airflow.providers.amazon.aws.sensors.dynamodb import DynamoDBValueSensor

pytestmark = pytest.mark.db_test


class TestDynamoDBValueSensor:
    def setup_method(self):
        self.table_name = "test_airflow"
        self.pk_name = "PK"
        self.pk_value = "PKTest"
        self.sk_name = "SK"
        self.sk_value = "SKTest"
        self.attribute_name = "Foo"
        self.attribute_value = "Bar"

        self.sensor_pk_sk = DynamoDBValueSensor(
            task_id="dynamodb_value_sensor",
            table_name=self.table_name,
            partition_key_name=self.pk_name,
            partition_key_value=self.pk_value,
            attribute_name=self.attribute_name,
            attribute_value=self.attribute_value,
            sort_key_name=self.sk_name,
            sort_key_value=self.sk_value,
        )

        self.sensor_pk = DynamoDBValueSensor(
            task_id="dynamodb_value_sensor",
            table_name=self.table_name,
            partition_key_name=self.pk_name,
            partition_key_value=self.pk_value,
            attribute_name=self.attribute_name,
            attribute_value=self.attribute_value,
        )

    @mock_aws
    def test_sensor_with_pk(self):
        hook = DynamoDBHook(table_name=self.table_name, table_keys=[self.pk_name])

        hook.conn.create_table(
            TableName=self.table_name,
            KeySchema=[{"AttributeName": self.pk_name, "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": self.pk_name, "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

        assert not self.sensor_pk.poke(None)

        items = [{self.pk_name: self.pk_value, self.attribute_name: self.attribute_value}]
        hook.write_batch_data(items)

        assert self.sensor_pk.poke(None)

    @mock_aws
    def test_sensor_with_pk_and_sk(self):
        hook = DynamoDBHook(table_name=self.table_name, table_keys=[self.pk_name, self.sk_name])

        hook.conn.create_table(
            TableName=self.table_name,
            KeySchema=[
                {"AttributeName": self.pk_name, "KeyType": "HASH"},
                {"AttributeName": self.sk_name, "KeyType": "RANGE"},
            ],
            AttributeDefinitions=[
                {"AttributeName": self.pk_name, "AttributeType": "S"},
                {"AttributeName": self.sk_name, "AttributeType": "S"},
            ],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

        assert not self.sensor_pk_sk.poke(None)

        items = [
            {
                self.pk_name: self.pk_value,
                self.sk_name: self.sk_value,
                self.attribute_name: self.attribute_value,
            }
        ]
        hook.write_batch_data(items)

        assert self.sensor_pk_sk.poke(None)

    @mock_aws
    def test_sensor_with_client_error(self):
        hook = DynamoDBHook(table_name=self.table_name, table_keys=[self.pk_name])

        hook.conn.create_table(
            TableName=self.table_name,
            KeySchema=[{"AttributeName": self.pk_name, "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": self.pk_name, "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

        items = [{self.pk_name: self.pk_value, self.attribute_name: self.attribute_value}]
        hook.write_batch_data(items)

        self.sensor_pk.partition_key_name = "no such key"
        assert self.sensor_pk.poke(None) is False


class TestDynamoDBMultipleValuesSensor:
    def setup_method(self):
        self.table_name = "test_airflow"
        self.pk_name = "PK"
        self.pk_value = "PKTest"
        self.sk_name = "SK"
        self.sk_value = "SKTest"
        self.attribute_name = "Foo"
        self.attribute_value = ["Bar1", "Bar2", "Bar3"]

        self.sensor = DynamoDBValueSensor(
            task_id="dynamodb_value_sensor",
            table_name=self.table_name,
            partition_key_name=self.pk_name,
            partition_key_value=self.pk_value,
            attribute_name=self.attribute_name,
            attribute_value=self.attribute_value,
            sort_key_name=self.sk_name,
            sort_key_value=self.sk_value,
        )
        self.sensor_pk = DynamoDBValueSensor(
            task_id="dynamodb_value_sensor",
            table_name=self.table_name,
            partition_key_name=self.pk_name,
            partition_key_value=self.pk_value,
            attribute_name=self.attribute_name,
            attribute_value=self.attribute_value,
        )

    def test_init(self):
        sensor = DynamoDBValueSensor(
            task_id="dynamodb_value_sensor_init",
            table_name=self.table_name,
            partition_key_name=self.pk_name,
            partition_key_value=self.pk_value,
            attribute_name=self.attribute_name,
            attribute_value=self.attribute_value,
            sort_key_name=self.sk_name,
            sort_key_value=self.sk_value,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="cn-north-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert sensor.hook.client_type is None
        assert sensor.hook.resource_type == "dynamodb"
        assert sensor.hook.aws_conn_id == "fake-conn-id"
        assert sensor.hook._region_name == "cn-north-1"
        assert sensor.hook._verify is False
        assert sensor.hook._config is not None
        assert sensor.hook._config.read_timeout == 42

        sensor = DynamoDBValueSensor(
            task_id="dynamodb_value_sensor_init",
            table_name=self.table_name,
            partition_key_name=self.pk_name,
            partition_key_value=self.pk_value,
            attribute_name=self.attribute_name,
            attribute_value=self.attribute_value,
            sort_key_name=self.sk_name,
            sort_key_value=self.sk_value,
        )
        assert sensor.hook.aws_conn_id == "aws_default"
        assert sensor.hook._region_name is None
        assert sensor.hook._verify is None
        assert sensor.hook._config is None

    @mock_aws
    def test_sensor_with_pk(self):
        hook = DynamoDBHook(table_name=self.table_name, table_keys=[self.pk_name])

        hook.conn.create_table(
            TableName=self.table_name,
            KeySchema=[{"AttributeName": self.pk_name, "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": self.pk_name, "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

        assert not self.sensor_pk.poke(None)

        items = [{self.pk_name: self.pk_value, self.attribute_name: self.attribute_value[1]}]
        hook.write_batch_data(items)

        assert self.sensor_pk.poke(None)

    @mock_aws
    def test_sensor_with_pk_and_sk(self):
        hook = DynamoDBHook(table_name=self.table_name, table_keys=[self.pk_name, self.sk_name])

        hook.conn.create_table(
            TableName=self.table_name,
            KeySchema=[
                {"AttributeName": self.pk_name, "KeyType": "HASH"},
                {"AttributeName": self.sk_name, "KeyType": "RANGE"},
            ],
            AttributeDefinitions=[
                {"AttributeName": self.pk_name, "AttributeType": "S"},
                {"AttributeName": self.sk_name, "AttributeType": "S"},
            ],
            ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
        )

        assert not self.sensor.poke(None)

        items = [
            {
                self.pk_name: self.pk_value,
                self.sk_name: self.sk_value,
                self.attribute_name: self.attribute_value[1],
            }
        ]
        hook.write_batch_data(items)

        assert self.sensor.poke(None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

from airflow.providers.amazon.aws.hooks.glue_databrew import GlueDataBrewHook

if TYPE_CHECKING:
    from unittest.mock import MagicMock


class TestGlueDataBrewHook:
    job_name = "test-databrew-job"
    runId = "test12345"

    @mock.patch.object(GlueDataBrewHook, "get_job_state")
    def test_get_job_state(self, get_job_state_mock: MagicMock):
        get_job_state_mock.return_value = "SUCCEEDED"
        hook = GlueDataBrewHook()
        result = hook.get_job_state(self.job_name, self.runId)
        assert result == "SUCCEEDED"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from botocore.exceptions import ClientError, WaiterError

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.dynamodb import DynamoDBHook
from airflow.providers.amazon.aws.transfers.s3_to_dynamodb import S3ToDynamoDBOperator

TASK_ID = "transfer_1"
BUCKET = "test-bucket"
S3_KEY_PREFIX = "test/test_data"
S3_KEY = "test/test_data_file_1.csv"
S3_CONN_ID = "aws_default"
DYNAMODB_TABLE_NAME = "test-table"
DYNAMODB_ATTRIBUTES = [
    {"AttributeName": "attribute_a", "AttributeType": "S"},
    {"AttributeName": "attribute_b", "AttributeType": "I"},
]
DYNAMODB_KEY_SCHEMA = [
    {"AttributeName": "attribute_a", "KeyType": "HASH"},
]

DYNAMODB_PROV_THROUGHPUT = {"ReadCapacityUnits": 123, "WriteCapacityUnits": 123}
SUCCESS_S3_RESPONSE = {
    "ImportTableDescription": {
        "ImportArn": "arn:aws:dynamodb:import",
        "ImportStatus": "IN_PROGRESS",
        "TableArn": "arn:aws:dynamodb:table",
        "TableId": "test-table",
        "ClientToken": "client",
    }
}
FAILURE_S3_RESPONSE = {
    "ImportTableDescription": {
        "ImportArn": "arn:aws:dynamodb:import",
        "ImportStatus": "FAILED",
        "TableArn": "arn:aws:dynamodb:table",
        "TableId": "test-table",
        "ClientToken": "client",
        "FailureCode": "300",
        "FailureMessage": "invalid csv format",
    }
}
IMPORT_TABLE_RESPONSE = {
    "S3BucketSource": {"S3Bucket": "test-bucket", "S3KeyPrefix": "test/test_data"},
    "InputFormat": "DYNAMODB_JSON",
    "TableCreationParameters": {
        "TableName": "test-table",
        "AttributeDefinitions": [
            {"AttributeName": "attribute_a", "AttributeType": "S"},
            {"AttributeName": "attribute_b", "AttributeType": "I"},
        ],
        "KeySchema": [{"AttributeName": "attribute_a", "KeyType": "HASH"}],
        "BillingMode": "PAY_PER_REQUEST",
        "ProvisionedThroughput": {"ReadCapacityUnits": 123, "WriteCapacityUnits": 123},
    },
}


@pytest.fixture
def new_table_op():
    return S3ToDynamoDBOperator(
        task_id=TASK_ID,
        s3_key=S3_KEY_PREFIX,
        s3_bucket=BUCKET,
        dynamodb_table_name=DYNAMODB_TABLE_NAME,
        dynamodb_attributes=DYNAMODB_ATTRIBUTES,
        dynamodb_key_schema=DYNAMODB_KEY_SCHEMA,
        aws_conn_id=S3_CONN_ID,
        import_table_creation_kwargs={"ProvisionedThroughput": DYNAMODB_PROV_THROUGHPUT},
    )


@pytest.fixture
def exist_table_op():
    return S3ToDynamoDBOperator(
        task_id=TASK_ID,
        s3_key=S3_KEY,
        dynamodb_key_schema=DYNAMODB_KEY_SCHEMA,
        s3_bucket=BUCKET,
        dynamodb_table_name=DYNAMODB_TABLE_NAME,
        use_existing_table=True,
        aws_conn_id=S3_CONN_ID,
    )


class TestS3ToDynamoDBOperator:
    @mock.patch.object(DynamoDBHook, "get_waiter")
    @mock.patch("botocore.client.BaseClient._make_api_call")
    def test_s3_to_dynamodb_new_table_wait_for_completion(self, mock_make_api_call, mock_wait, new_table_op):
        mock_make_api_call.return_value = SUCCESS_S3_RESPONSE

        res = new_table_op.execute(None)

        mock_make_api_call.assert_called_once_with("ImportTable", IMPORT_TABLE_RESPONSE)
        mock_wait.assert_called_once_with("import_table")
        mock_wait.return_value.wait.assert_called_once_with(
            ImportArn="arn:aws:dynamodb:import", WaiterConfig={"Delay": 30, "MaxAttempts": 240}
        )
        assert res == "arn:aws:dynamodb:import"

    @pytest.mark.parametrize(
        "delete_on_error",
        [
            pytest.param(
                True,
                id="delete-on-error",
            ),
            pytest.param(
                False,
                id="no-delete-on-error",
            ),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.transfers.s3_to_dynamodb.DynamoDBHook")
    def test_s3_to_dynamodb_new_table_delete_on_error(self, mock_hook, new_table_op, delete_on_error):
        mock_wait = mock.Mock()
        mock_wait.side_effect = WaiterError(name="NetworkError", reason="unit test error", last_response={})
        mock_hook.return_value.get_waiter.return_value.wait = mock_wait
        new_table_op.delete_on_error = delete_on_error
        mock_hook.return_value.get_import_status.return_value = "FAILED", "400", "General error"

        with pytest.raises(AirflowException):
            new_table_op.execute(None)

        if delete_on_error:
            mock_hook.return_value.client.delete_table.assert_called_once_with(TableName="test-table")
        else:
            mock_hook.return_value.client.delete_table.assert_not_called()

    @mock.patch("botocore.client.BaseClient._make_api_call")
    def test_s3_to_dynamodb_new_table_no_wait(self, mock_make_api_call):
        mock_make_api_call.return_value = SUCCESS_S3_RESPONSE
        op = S3ToDynamoDBOperator(
            task_id=TASK_ID,
            s3_key=S3_KEY_PREFIX,
            s3_bucket=BUCKET,
            dynamodb_table_name=DYNAMODB_TABLE_NAME,
            dynamodb_attributes=DYNAMODB_ATTRIBUTES,
            dynamodb_key_schema=DYNAMODB_KEY_SCHEMA,
            aws_conn_id=S3_CONN_ID,
            import_table_creation_kwargs={"ProvisionedThroughput": DYNAMODB_PROV_THROUGHPUT},
            wait_for_completion=False,
        )
        res = op.execute(None)

        mock_make_api_call.assert_called_once_with("ImportTable", IMPORT_TABLE_RESPONSE)
        assert res == "arn:aws:dynamodb:import"

    @mock.patch("botocore.client.BaseClient._make_api_call")
    def test_s3_to_dynamodb_new_table_client_error(self, mock_make_api_call, new_table_op):
        mock_make_api_call.side_effect = ClientError(
            error_response={"Error": {"Message": "Error message", "Code": "GeneralException"}},
            operation_name="UnitTest",
        )
        with pytest.raises(AirflowException) as excinfo:
            new_table_op.execute(None)
        assert "S3 load into DynamoDB table failed with error" in str(excinfo.value), (
            "Exception message not passed correctly"
        )

    @mock.patch("botocore.client.BaseClient._make_api_call")
    def test_s3_to_dynamodb_new_table_job_startup_error(self, mock_make_api_call, new_table_op):
        mock_make_api_call.return_value = FAILURE_S3_RESPONSE
        exp_err_msg = "S3 into Dynamodb job creation failed. Code: 300. Failure: invalid csv format"
        with pytest.raises(AirflowException) as excinfo:
            new_table_op.execute(None)
        assert str(excinfo.value) == exp_err_msg, "Exception message not passed correctly"

    @mock.patch(
        "airflow.providers.amazon.aws.transfers.s3_to_dynamodb.S3ToDynamoDBOperator._load_into_new_table"
    )
    @mock.patch.object(DynamoDBHook, "get_conn")
    def test_s3_to_dynamodb_existing_table(self, mock_get_conn, new_table_load_mock, exist_table_op):
        response = [
            {
                "Items": [
                    {"Date": {"N": "54675846"}, "Message": {"S": "Message1"}, "_id": {"S": "1"}},
                    {"Date": {"N": "54675847"}, "Message": {"S": "Message2"}, "_id": {"S": "2"}},
                    {"Date": {"N": "54675857"}, "Message": {"S": "Message3"}, "_id": {"S": "4"}},
                ]
            }
        ]
        batch_writer_calls = [mock.call(Item=item) for item in response[0]["Items"]]
        mock_paginator = mock.Mock()
        mock_paginator.paginate.return_value = response

        mock_conn = mock.MagicMock()
        mock_client = mock.Mock()
        mock_put_item = mock.Mock()

        mock_client.get_paginator.return_value = mock_paginator
        mock_conn.meta.client = mock_client
        mock_conn.Table.return_value.batch_writer.return_value.__enter__.return_value.put_item = mock_put_item
        mock_conn.Table.return_value.table_arn = "arn:aws:dynamodb"
        mock_get_conn.return_value = mock_conn

        res = exist_table_op.execute(None)

        new_table_load_mock.assert_called_once_with(
            table_name=exist_table_op.tmp_table_name, delete_on_error=False
        )
        mock_client.get_paginator.assert_called_once_with("scan")
        mock_client.get_paginator.return_value.paginate.assert_called_once_with(
            TableName=exist_table_op.tmp_table_name,
            Select="ALL_ATTRIBUTES",
            ReturnConsumedCapacity="NONE",
            ConsistentRead=True,
        )
        mock_conn.Table.assert_called_with("test-table")
        mock_conn.Table.return_value.batch_writer.assert_called_once_with(overwrite_by_pkeys=["attribute_a"])
        mock_put_item.assert_has_calls(batch_writer_calls)
        mock_client.delete_table.assert_called_once_with(TableName=exist_table_op.tmp_table_name)
        assert res == "arn:aws:dynamodb"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import gzip
import io
from unittest import mock

import pandas as pd
import pytest

from airflow.exceptions import AirflowException
from airflow.models import Connection
from airflow.providers.amazon.aws.transfers.sql_to_s3 import SqlToS3Operator


class TestSqlToS3Operator:
    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_execute_csv(self, mock_s3_hook):
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            pd_kwargs={"index": False, "header": False},
            dag=None,
        )
        op._get_hook = mock_dbapi_hook
        op.execute(None)

        mock_s3_hook.assert_called_once_with(aws_conn_id="aws_conn_id", verify=None)
        get_df_mock.assert_called_once_with(sql=query, parameters=None, df_type="pandas")
        file_obj = mock_s3_hook.return_value.load_file_obj.call_args[1]["file_obj"]
        assert isinstance(file_obj, io.BytesIO)
        mock_s3_hook.return_value.load_file_obj.assert_called_once_with(
            file_obj=file_obj, key=s3_key, bucket_name=s3_bucket, replace=True
        )

    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_execute_parquet(self, mock_s3_hook):
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])

        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            file_format="parquet",
            replace=True,
            dag=None,
        )
        op._get_hook = mock_dbapi_hook
        op.execute(None)

        mock_s3_hook.assert_called_once_with(aws_conn_id="aws_conn_id", verify=None)
        get_df_mock.assert_called_once_with(sql=query, parameters=None, df_type="pandas")
        file_obj = mock_s3_hook.return_value.load_file_obj.call_args[1]["file_obj"]
        assert isinstance(file_obj, io.BytesIO)
        mock_s3_hook.return_value.load_file_obj.assert_called_once_with(
            file_obj=file_obj, key=s3_key, bucket_name=s3_bucket, replace=True
        )

    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_execute_json(self, mock_s3_hook):
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            file_format="json",
            replace=True,
            pd_kwargs={"date_format": "iso", "lines": True, "orient": "records"},
            dag=None,
        )
        op._get_hook = mock_dbapi_hook
        op.execute(None)

        mock_s3_hook.assert_called_once_with(aws_conn_id="aws_conn_id", verify=None)
        get_df_mock.assert_called_once_with(sql=query, parameters=None, df_type="pandas")
        file_obj = mock_s3_hook.return_value.load_file_obj.call_args[1]["file_obj"]
        assert isinstance(file_obj, io.BytesIO)
        mock_s3_hook.return_value.load_file_obj.assert_called_once_with(
            file_obj=file_obj, key=s3_key, bucket_name=s3_bucket, replace=True
        )

    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_execute_gzip_with_bytesio(self, mock_s3_hook):
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key.csv.gz"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            pd_kwargs={"index": False, "compression": "gzip"},
            dag=None,
        )
        op._get_hook = mock_dbapi_hook
        op.execute(None)

        mock_s3_hook.assert_called_once_with(aws_conn_id="aws_conn_id", verify=None)
        get_df_mock.assert_called_once_with(sql=query, parameters=None, df_type="pandas")
        file_obj = mock_s3_hook.return_value.load_file_obj.call_args[1]["file_obj"]
        assert isinstance(file_obj, io.BytesIO)
        mock_s3_hook.return_value.load_file_obj.assert_called_once_with(
            file_obj=file_obj, key=s3_key, bucket_name=s3_bucket, replace=True
        )

        file_obj.seek(0)
        with gzip.GzipFile(fileobj=file_obj, mode="rb") as gz:
            decompressed_buf = io.BytesIO(gz.read())
        decompressed_buf.seek(0)
        read_df = pd.read_csv(decompressed_buf, dtype={"a": str, "b": str})
        assert read_df.equals(test_df)

    @pytest.mark.parametrize(
        "params",
        [
            pytest.param({"file_format": "csv", "null_string_result": None}, id="with-csv"),
            pytest.param({"file_format": "parquet", "null_string_result": "None"}, id="with-parquet"),
        ],
    )
    def test_fix_dtypes(self, params):
        op = SqlToS3Operator(
            query="query",
            s3_bucket="s3_bucket",
            s3_key="s3_key",
            task_id="task_id",
            sql_conn_id="mysql_conn_id",
        )
        dirty_df = pd.DataFrame({"strings": ["a", "b", None], "ints": [1, 2, None]})
        op._fix_dtypes(df=dirty_df, file_format=params["file_format"])
        assert dirty_df["strings"].values[2] == params["null_string_result"]
        assert dirty_df["ints"].dtype.kind == "i"

    def test_invalid_file_format(self):
        with pytest.raises(AirflowException):
            SqlToS3Operator(
                query="query",
                s3_bucket="bucket",
                s3_key="key",
                sql_conn_id="mysql_conn_id",
                task_id="task_id",
                file_format="invalid_format",
                dag=None,
            )

    def test_with_groupby_kwarg(self):
        """
        Test operator when the groupby_kwargs is specified
        """
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            pd_kwargs={"index": False, "header": False},
            groupby_kwargs={"by": "Team"},
            dag=None,
        )
        example = {
            "Team": ["Australia", "Australia", "India", "India"],
            "Player": ["Ricky", "David Warner", "Virat Kohli", "Rohit Sharma"],
            "Runs": [345, 490, 672, 560],
        }

        df = pd.DataFrame(example)
        data = []
        for group_name, df in op._partition_dataframe(df):
            data.append((group_name, df))
        data.sort(key=lambda d: d[0])
        team, df = data[0]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["Australia", "Australia"],
                    "Player": ["Ricky", "David Warner"],
                    "Runs": [345, 490],
                }
            )
        )
        team, df = data[1]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["India", "India"],
                    "Player": ["Virat Kohli", "Rohit Sharma"],
                    "Runs": [672, 560],
                }
            )
        )

    def test_without_groupby_kwarg(self):
        """
        Test operator when the groupby_kwargs is not specified
        """
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            pd_kwargs={"index": False, "header": False},
            dag=None,
        )
        example = {
            "Team": ["Australia", "Australia", "India", "India"],
            "Player": ["Ricky", "David Warner", "Virat Kohli", "Rohit Sharma"],
            "Runs": [345, 490, 672, 560],
        }

        df = pd.DataFrame(example)
        data = []
        for group_name, df in op._partition_dataframe(df):
            data.append((group_name, df))

        assert len(data) == 1
        team, df = data[0]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["Australia", "Australia", "India", "India"],
                    "Player": ["Ricky", "David Warner", "Virat Kohli", "Rohit Sharma"],
                    "Runs": [345, 490, 672, 560],
                }
            )
        )

    def test_with_max_rows_per_file(self):
        """
        Test operator when the max_rows_per_file is specified
        """
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            pd_kwargs={"index": False, "header": False},
            max_rows_per_file=3,
            dag=None,
        )
        example = {
            "Team": ["Australia", "Australia", "India", "India"],
            "Player": ["Ricky", "David Warner", "Virat Kohli", "Rohit Sharma"],
            "Runs": [345, 490, 672, 560],
        }

        df = pd.DataFrame(example)
        data = []
        for group_name, df in op._partition_dataframe(df):
            data.append((group_name, df))
        data.sort(key=lambda d: d[0])
        team, df = data[0]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["Australia", "Australia", "India"],
                    "Player": ["Ricky", "David Warner", "Virat Kohli"],
                    "Runs": [345, 490, 672],
                }
            )
        )
        team, df = data[1]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["India"],
                    "Player": ["Rohit Sharma"],
                    "Runs": [560],
                }
            )
        )

    @mock.patch("airflow.providers.common.sql.operators.sql.BaseHook.get_connection")
    def test_hook_params(self, mock_get_conn):
        mock_get_conn.return_value = Connection(conn_id="postgres_test", conn_type="postgres")
        op = SqlToS3Operator(
            query="query",
            s3_bucket="bucket",
            s3_key="key",
            sql_conn_id="postgres_test",
            task_id="task_id",
            sql_hook_params={
                "log_sql": False,
            },
            dag=None,
        )
        hook = op._get_hook()
        assert hook.log_sql == op.sql_hook_params["log_sql"]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module allows you to transfer data from any Google API endpoint into a S3 Bucket."""

from __future__ import annotations

import json
import sys
from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.models import BaseOperator
from airflow.models.xcom import MAX_XCOM_SIZE, XCOM_RETURN_KEY
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.google.common.hooks.discovery_api import GoogleDiscoveryApiHook

if TYPE_CHECKING:
    try:
        from airflow.sdk.types import RuntimeTaskInstanceProtocol
    except ImportError:
        from airflow.models import TaskInstance as RuntimeTaskInstanceProtocol  # type: ignore[assignment]
    from airflow.utils.context import Context


class GoogleApiToS3Operator(BaseOperator):
    """
    Basic class for transferring data from a Google API endpoint into a S3 Bucket.

    This discovery-based operator use
    :class:`~airflow.providers.google.common.hooks.discovery_api.GoogleDiscoveryApiHook` to communicate
    with Google Services via the
    `Google API Python Client <https://github.com/googleapis/google-api-python-client>`__.
    Please note that this library is in maintenance mode hence it won't fully support Google Cloud in
    the future.
    Therefore it is recommended that you use the custom Google Cloud Service Operators for working
    with the Google Cloud Platform.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleApiToS3Operator`

    :param google_api_service_name: The specific API service that is being requested.
    :param google_api_service_version: The version of the API that is being requested.
    :param google_api_endpoint_path: The client libraries path to the api call's executing method.
        For example: 'analyticsreporting.reports.batchGet'

        .. note:: See https://developers.google.com/apis-explorer
            for more information on which methods are available.

    :param google_api_endpoint_params: The params to control the corresponding endpoint result.
    :param s3_destination_key: The url where to put the data retrieved from the endpoint in S3.

        .. note See https://docs.aws.amazon.com/AmazonS3/latest/userguide/access-bucket-intro.html
            for valid url formats.

    :param google_api_response_via_xcom: Can be set to expose the google api response to xcom.
    :param google_api_endpoint_params_via_xcom: If set to a value this value will be used as a key
        for pulling from xcom and updating the google api endpoint params.
    :param google_api_endpoint_params_via_xcom_task_ids: Task ids to filter xcom by.
    :param google_api_pagination: If set to True Pagination will be enabled for this request
        to retrieve all data.

        .. note:: This means the response will be a list of responses.

    :param google_api_num_retries: Define the number of retries for the Google API requests being made
        if it fails.
    :param s3_overwrite: Specifies whether the s3 file will be overwritten if exists.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param aws_conn_id: The connection id specifying the authentication information for the S3 Bucket.
    :param google_impersonation_chain: Optional Google service account to impersonate using
        short-term credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "google_api_endpoint_params",
        "s3_destination_key",
        "google_impersonation_chain",
        "gcp_conn_id",
    )
    template_ext: Sequence[str] = ()
    ui_color = "#cc181e"

    def __init__(
        self,
        *,
        google_api_service_name: str,
        google_api_service_version: str,
        google_api_endpoint_path: str,
        google_api_endpoint_params: dict,
        s3_destination_key: str,
        google_api_response_via_xcom: str | None = None,
        google_api_endpoint_params_via_xcom: str | None = None,
        google_api_endpoint_params_via_xcom_task_ids: str | None = None,
        google_api_pagination: bool = False,
        google_api_num_retries: int = 0,
        s3_overwrite: bool = False,
        gcp_conn_id: str = "google_cloud_default",
        aws_conn_id: str | None = "aws_default",
        google_impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.google_api_service_name = google_api_service_name
        self.google_api_service_version = google_api_service_version
        self.google_api_endpoint_path = google_api_endpoint_path
        self.google_api_endpoint_params = google_api_endpoint_params
        self.s3_destination_key = s3_destination_key
        self.google_api_response_via_xcom = google_api_response_via_xcom
        self.google_api_endpoint_params_via_xcom = google_api_endpoint_params_via_xcom
        self.google_api_endpoint_params_via_xcom_task_ids = google_api_endpoint_params_via_xcom_task_ids
        self.google_api_pagination = google_api_pagination
        self.google_api_num_retries = google_api_num_retries
        self.s3_overwrite = s3_overwrite
        self.gcp_conn_id = gcp_conn_id
        self.aws_conn_id = aws_conn_id
        self.google_impersonation_chain = google_impersonation_chain

    def execute(self, context: Context) -> None:
        """
        Transfers Google APIs json data to S3.

        :param context: The context that is being provided when executing.
        """
        self.log.info("Transferring data from %s to s3", self.google_api_service_name)

        if self.google_api_endpoint_params_via_xcom:
            self._update_google_api_endpoint_params_via_xcom(context["task_instance"])

        data = self._retrieve_data_from_google_api()

        self._load_data_to_s3(data)

        if self.google_api_response_via_xcom:
            self._expose_google_api_response_via_xcom(context["task_instance"], data)

    def _retrieve_data_from_google_api(self) -> dict:
        google_discovery_api_hook = GoogleDiscoveryApiHook(
            gcp_conn_id=self.gcp_conn_id,
            api_service_name=self.google_api_service_name,
            api_version=self.google_api_service_version,
            impersonation_chain=self.google_impersonation_chain,
        )
        return google_discovery_api_hook.query(
            endpoint=self.google_api_endpoint_path,
            data=self.google_api_endpoint_params,
            paginate=self.google_api_pagination,
            num_retries=self.google_api_num_retries,
        )

    def _load_data_to_s3(self, data: dict) -> None:
        s3_hook = S3Hook(aws_conn_id=self.aws_conn_id)
        s3_hook.load_string(
            string_data=json.dumps(data),
            bucket_name=S3Hook.parse_s3_url(self.s3_destination_key)[0],
            key=S3Hook.parse_s3_url(self.s3_destination_key)[1],
            replace=self.s3_overwrite,
        )

    def _update_google_api_endpoint_params_via_xcom(self, task_instance: RuntimeTaskInstanceProtocol) -> None:
        if self.google_api_endpoint_params_via_xcom:
            google_api_endpoint_params = task_instance.xcom_pull(
                task_ids=self.google_api_endpoint_params_via_xcom_task_ids,
                key=self.google_api_endpoint_params_via_xcom,
            )
            self.google_api_endpoint_params.update(google_api_endpoint_params)

    def _expose_google_api_response_via_xcom(
        self, task_instance: RuntimeTaskInstanceProtocol, data: dict
    ) -> None:
        if sys.getsizeof(data) < MAX_XCOM_SIZE:
            task_instance.xcom_push(key=self.google_api_response_via_xcom or XCOM_RETURN_KEY, value=data)
        else:
            raise RuntimeError("The size of the downloaded data is too large to push to XCom!")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import time
import warnings
from functools import cached_property
from typing import Any

from botocore.exceptions import ClientError

from airflow.exceptions import AirflowException, AirflowProviderDeprecationWarning
from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.amazon.aws.hooks.logs import AwsLogsHook

DEFAULT_LOG_SUFFIX = "output"
ERROR_LOG_SUFFIX = "error"


class GlueJobHook(AwsBaseHook):
    """
    Interact with AWS Glue.

    Provide thick wrapper around :external+boto3:py:class:`boto3.client("glue") <Glue.Client>`.

    :param s3_bucket: S3 bucket where logs and local etl script will be uploaded
    :param job_name: unique job name per AWS account
    :param desc: job description
    :param concurrent_run_limit: The maximum number of concurrent runs allowed for a job
    :param script_location: path to etl script on s3
    :param retry_limit: Maximum number of times to retry this job if it fails
    :param num_of_dpus: Number of AWS Glue DPUs to allocate to this Job
    :param region_name: aws region name (example: us-east-1)
    :param iam_role_name: AWS IAM Role for Glue Job Execution. If set `iam_role_arn` must equal None.
    :param iam_role_arn: AWS IAM Role ARN for Glue Job Execution, If set `iam_role_name` must equal None.
    :param create_job_kwargs: Extra arguments for Glue Job Creation
    :param update_config: Update job configuration on Glue (default: False)

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    class LogContinuationTokens:
        """Used to hold the continuation tokens when reading logs from both streams Glue Jobs write to."""

        def __init__(self):
            self.output_stream_continuation: str | None = None
            self.error_stream_continuation: str | None = None

    def __init__(
        self,
        s3_bucket: str | None = None,
        job_name: str | None = None,
        desc: str | None = None,
        concurrent_run_limit: int = 1,
        script_location: str | None = None,
        retry_limit: int = 0,
        num_of_dpus: int | float | None = None,
        iam_role_name: str | None = None,
        iam_role_arn: str | None = None,
        create_job_kwargs: dict | None = None,
        update_config: bool = False,
        job_poll_interval: int | float = 6,
        *args,
        **kwargs,
    ):
        self.job_name = job_name
        self.desc = desc
        self.concurrent_run_limit = concurrent_run_limit
        self.script_location = script_location
        self.retry_limit = retry_limit
        self.s3_bucket = s3_bucket
        self.role_name = iam_role_name
        self.role_arn = iam_role_arn
        self.s3_glue_logs = "logs/glue-logs/"
        self.create_job_kwargs = create_job_kwargs or {}
        self.update_config = update_config
        self.job_poll_interval = job_poll_interval

        worker_type_exists = "WorkerType" in self.create_job_kwargs
        num_workers_exists = "NumberOfWorkers" in self.create_job_kwargs

        if self.role_arn and self.role_name:
            raise ValueError("Cannot set iam_role_arn and iam_role_name simultaneously")
        if worker_type_exists and num_workers_exists:
            if num_of_dpus is not None:
                raise ValueError("Cannot specify num_of_dpus with custom WorkerType")
        elif not worker_type_exists and num_workers_exists:
            raise ValueError("Need to specify custom WorkerType when specifying NumberOfWorkers")
        elif worker_type_exists and not num_workers_exists:
            raise ValueError("Need to specify NumberOfWorkers when specifying custom WorkerType")
        elif num_of_dpus is None:
            self.num_of_dpus: int | float = 10
        else:
            self.num_of_dpus = num_of_dpus

        kwargs["client_type"] = "glue"
        super().__init__(*args, **kwargs)

    def create_glue_job_config(self) -> dict:
        default_command = {
            "Name": "glueetl",
            "ScriptLocation": self.script_location,
        }
        command = self.create_job_kwargs.pop("Command", default_command)
        if not self.role_arn:
            execution_role = self.get_iam_execution_role()
            role_arn = execution_role["Role"]["Arn"]
        else:
            role_arn = self.role_arn

        config = {
            "Name": self.job_name,
            "Description": self.desc,
            "Role": role_arn,
            "ExecutionProperty": {"MaxConcurrentRuns": self.concurrent_run_limit},
            "Command": command,
            "MaxRetries": self.retry_limit,
            **self.create_job_kwargs,
        }

        if hasattr(self, "num_of_dpus"):
            config["MaxCapacity"] = self.num_of_dpus

        if self.s3_bucket is not None:
            config["LogUri"] = f"s3://{self.s3_bucket}/{self.s3_glue_logs}{self.job_name}"

        return config

    def describe_jobs(self) -> list:
        """
        Get list of Jobs.

        .. seealso::
            - :external+boto3:py:meth:`Glue.Client.get_jobs`
        """
        return self.conn.get_jobs()

    def list_jobs(self) -> list:
        """
        Get list of Jobs.

        .. deprecated::
            - Use :meth:`describe_jobs` instead.
        """
        warnings.warn(
            "The method `list_jobs` is deprecated. Use the method `describe_jobs` instead.",
            AirflowProviderDeprecationWarning,
            stacklevel=2,
        )
        return self.describe_jobs()

    def get_iam_execution_role(self) -> dict:
        try:
            iam_client = self.get_session(region_name=self.region_name).client(
                "iam", endpoint_url=self.conn_config.endpoint_url, config=self.config, verify=self.verify
            )
            glue_execution_role = iam_client.get_role(RoleName=self.role_name)
            self.log.info("Iam Role Name: %s", self.role_name)
            return glue_execution_role
        except Exception as general_error:
            self.log.error("Failed to create aws glue job, error: %s", general_error)
            raise

    def initialize_job(
        self,
        script_arguments: dict | None = None,
        run_kwargs: dict | None = None,
    ) -> dict[str, str]:
        """
        Initialize connection with AWS Glue to run job.

        .. seealso::
            - :external+boto3:py:meth:`Glue.Client.start_job_run`
        """
        script_arguments = script_arguments or {}
        run_kwargs = run_kwargs or {}

        try:
            if self.update_config:
                job_name = self.create_or_update_glue_job()
            else:
                job_name = self.get_or_create_glue_job()

            return self.conn.start_job_run(JobName=job_name, Arguments=script_arguments, **run_kwargs)
        except Exception as general_error:
            self.log.error("Failed to run aws glue job, error: %s", general_error)
            raise

    def get_job_state(self, job_name: str, run_id: str) -> str:
        """
        Get state of the Glue job; the job state can be running, finished, failed, stopped or timeout.

        .. seealso::
            - :external+boto3:py:meth:`Glue.Client.get_job_run`

        :param job_name: unique job name per AWS account
        :param run_id: The job-run ID of the predecessor job run
        :return: State of the Glue job
        """
        job_run = self.conn.get_job_run(JobName=job_name, RunId=run_id, PredecessorsIncluded=True)
        return job_run["JobRun"]["JobRunState"]

    async def async_get_job_state(self, job_name: str, run_id: str) -> str:
        """
        Get state of the Glue job; the job state can be running, finished, failed, stopped or timeout.

        The async version of get_job_state.
        """
        async with await self.get_async_conn() as client:
            job_run = await client.get_job_run(JobName=job_name, RunId=run_id)
        return job_run["JobRun"]["JobRunState"]

    @cached_property
    def logs_hook(self):
        """Returns an AwsLogsHook instantiated with the parameters of the GlueJobHook."""
        return AwsLogsHook(
            aws_conn_id=self.aws_conn_id, region_name=self.region_name, verify=self.verify, config=self.config
        )

    def print_job_logs(
        self,
        job_name: str,
        run_id: str,
        continuation_tokens: LogContinuationTokens,
    ):
        """
        Print the latest job logs to the Airflow task log and updates the continuation tokens.

        :param continuation_tokens: the tokens where to resume from when reading logs.
            The object gets updated with the new tokens by this method.
        """
        log_client = self.logs_hook.get_conn()
        paginator = log_client.get_paginator("filter_log_events")
        job_run = self.conn.get_job_run(JobName=job_name, RunId=run_id)["JobRun"]
        # StartTime needs to be an int and is Epoch time in milliseconds
        start_time = int(job_run["StartedOn"].timestamp() * 1000)

        def display_logs_from(log_group: str, continuation_token: str | None) -> str | None:
            """Mutualize iteration over the 2 different log streams glue jobs write to."""
            fetched_logs = []
            next_token = continuation_token
            try:
                for response in paginator.paginate(
                    logGroupName=log_group,
                    logStreamNames=[run_id],
                    startTime=start_time,
                    PaginationConfig={"StartingToken": continuation_token},
                ):
                    fetched_logs.extend([event["message"] for event in response["events"]])
                    # if the response is empty there is no nextToken in it
                    next_token = response.get("nextToken") or next_token
            except ClientError as e:
                if e.response["Error"]["Code"] == "ResourceNotFoundException":
                    # we land here when the log groups/streams don't exist yet
                    self.log.warning(
                        "No new Glue driver logs so far.\n"
                        "If this persists, check the CloudWatch dashboard at: %r.",
                        f"https://{self.conn_region_name}.console.aws.amazon.com/cloudwatch/home",
                    )
                else:
                    raise

            if len(fetched_logs):
                # Add a tab to indent those logs and distinguish them from airflow logs.
                # Log lines returned already contain a newline character at the end.
                messages = "\t".join(fetched_logs)
                self.log.info("Glue Job Run %s Logs:\n\t%s", log_group, messages)
            else:
                self.log.info("No new log from the Glue Job in %s", log_group)
            return next_token

        log_group_prefix = job_run["LogGroupName"]
        log_group_default = f"{log_group_prefix}/{DEFAULT_LOG_SUFFIX}"
        log_group_error = f"{log_group_prefix}/{ERROR_LOG_SUFFIX}"
        # one would think that the error log group would contain only errors, but it actually contains
        # a lot of interesting logs too, so it's valuable to have both
        continuation_tokens.output_stream_continuation = display_logs_from(
            log_group_default, continuation_tokens.output_stream_continuation
        )
        continuation_tokens.error_stream_continuation = display_logs_from(
            log_group_error, continuation_tokens.error_stream_continuation
        )

    def job_completion(
        self, job_name: str, run_id: str, verbose: bool = False, sleep_before_return: int = 0
    ) -> dict[str, str]:
        """
        Wait until Glue job with job_name finishes; return final state if finished or raises AirflowException.

        :param job_name: unique job name per AWS account
        :param run_id: The job-run ID of the predecessor job run
        :param verbose: If True, more Glue Job Run logs show in the Airflow Task Logs.  (default: False)
        :param sleep_before_return: time in seconds to wait before returning final status.
        :return: Dict of JobRunState and JobRunId
        """
        next_log_tokens = self.LogContinuationTokens()
        while True:
            job_run_state = self.get_job_state(job_name, run_id)
            ret = self._handle_state(job_run_state, job_name, run_id, verbose, next_log_tokens)
            if ret:
                time.sleep(sleep_before_return)
                return ret
            time.sleep(self.job_poll_interval)

    async def async_job_completion(self, job_name: str, run_id: str, verbose: bool = False) -> dict[str, str]:
        """
        Wait until Glue job with job_name finishes; return final state if finished or raises AirflowException.

        :param job_name: unique job name per AWS account
        :param run_id: The job-run ID of the predecessor job run
        :param verbose: If True, more Glue Job Run logs show in the Airflow Task Logs.  (default: False)
        :return: Dict of JobRunState and JobRunId
        """
        next_log_tokens = self.LogContinuationTokens()
        while True:
            job_run_state = await self.async_get_job_state(job_name, run_id)
            ret = self._handle_state(job_run_state, job_name, run_id, verbose, next_log_tokens)
            if ret:
                return ret
            await asyncio.sleep(self.job_poll_interval)

    def _handle_state(
        self,
        state: str,
        job_name: str,
        run_id: str,
        verbose: bool,
        next_log_tokens: GlueJobHook.LogContinuationTokens,
    ) -> dict | None:
        """Process Glue Job state while polling; used by both sync and async methods."""
        failed_states = ["FAILED", "TIMEOUT"]
        finished_states = ["SUCCEEDED", "STOPPED"]

        if verbose:
            self.print_job_logs(
                job_name=job_name,
                run_id=run_id,
                continuation_tokens=next_log_tokens,
            )

        if state in finished_states:
            self.log.info("Exiting Job %s Run State: %s", run_id, state)
            return {"JobRunState": state, "JobRunId": run_id}
        if state in failed_states:
            job_error_message = f"Exiting Job {run_id} Run State: {state}"
            self.log.info(job_error_message)
            raise AirflowException(job_error_message)
        self.log.info(
            "Polling for AWS Glue Job %s current run state with status %s",
            job_name,
            state,
        )
        return None

    def has_job(self, job_name) -> bool:
        """
        Check if the job already exists.

        .. seealso::
            - :external+boto3:py:meth:`Glue.Client.get_job`

        :param job_name: unique job name per AWS account
        :return: Returns True if the job already exists and False if not.
        """
        self.log.info("Checking if job already exists: %s", job_name)

        try:
            self.conn.get_job(JobName=job_name)
            return True
        except self.conn.exceptions.EntityNotFoundException:
            return False

    def update_job(self, **job_kwargs) -> bool:
        """
        Update job configurations.

        .. seealso::
            - :external+boto3:py:meth:`Glue.Client.update_job`

        :param job_kwargs: Keyword args that define the configurations used for the job
        :return: True if job was updated and false otherwise
        """
        job_name = job_kwargs.pop("Name")
        current_job = self.conn.get_job(JobName=job_name)["Job"]

        update_config = {
            key: value for key, value in job_kwargs.items() if current_job.get(key) != job_kwargs[key]
        }
        if update_config != {}:
            self.log.info("Updating job: %s", job_name)
            self.conn.update_job(JobName=job_name, JobUpdate=job_kwargs)
            self.log.info("Updated configurations: %s", update_config)
            return True
        return False

    def get_or_create_glue_job(self) -> str | None:
        """
        Get (or creates) and returns the Job name.

        .. seealso::
            - :external+boto3:py:meth:`Glue.Client.create_job`

        :return:Name of the Job
        """
        if self.has_job(self.job_name):
            return self.job_name

        config = self.create_glue_job_config()
        self.log.info("Creating job: %s", self.job_name)
        self.conn.create_job(**config)

        return self.job_name

    def create_or_update_glue_job(self) -> str | None:
        """
        Create (or update) and return the Job name.

        .. seealso::
            - :external+boto3:py:meth:`Glue.Client.update_job`
            - :external+boto3:py:meth:`Glue.Client.create_job`

        :return:Name of the Job
        """
        config = self.create_glue_job_config()

        if self.has_job(self.job_name):
            self.update_job(**config)
        else:
            self.log.info("Creating job: %s", self.job_name)
            self.conn.create_job(**config)

        return self.job_name


class GlueDataQualityHook(AwsBaseHook):
    """
    Interact with AWS Glue Data Quality.

    Provide thick wrapper around :external+boto3:py:class:`boto3.client("glue") <Glue.Client>`.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        kwargs["client_type"] = "glue"
        super().__init__(*args, **kwargs)

    def has_data_quality_ruleset(self, name: str) -> bool:
        try:
            self.conn.get_data_quality_ruleset(Name=name)
            return True
        except self.conn.exceptions.EntityNotFoundException:
            return False

    def _log_results(self, result: dict[str, Any]) -> None:
        """
        Print the outcome of evaluation run, An evaluation run can involve multiple rulesets evaluated against a data source (Glue table).

        Name    Description                                     Result        EvaluatedMetrics                                                                    EvaluationMessage
        Rule_1    RowCount between 150000 and 600000             PASS        {'Dataset.*.RowCount': 300000.0}                                                       NaN
        Rule_2    IsComplete "marketplace"                       PASS        {'Column.marketplace.Completeness': 1.0}                                               NaN
        Rule_3    ColumnLength "marketplace" between 1 and 2     FAIL        {'Column.marketplace.MaximumLength': 9.0, 'Column.marketplace.MinimumLength': 3.0}     Value: 9.0 does not meet the constraint requirement!

        """
        import pandas as pd

        pd.set_option("display.max_rows", None)
        pd.set_option("display.max_columns", None)
        pd.set_option("display.width", None)
        pd.set_option("display.max_colwidth", None)

        self.log.info(
            "AWS Glue data quality ruleset evaluation result for RulesetName: %s RulesetEvaluationRunId: %s Score: %s",
            result.get("RulesetName"),
            result.get("RulesetEvaluationRunId"),
            result.get("Score"),
        )

        rule_results = result["RuleResults"]
        rule_results_df = pd.DataFrame(rule_results)
        self.log.info(rule_results_df)

    def get_evaluation_run_results(self, run_id: str) -> dict[str, Any]:
        response = self.conn.get_data_quality_ruleset_evaluation_run(RunId=run_id)

        return self.conn.batch_get_data_quality_result(ResultIds=response["ResultIds"])

    def validate_evaluation_run_results(
        self, evaluation_run_id: str, show_results: bool = True, verify_result_status: bool = True
    ) -> None:
        results = self.get_evaluation_run_results(evaluation_run_id)
        total_failed_rules = 0

        if results.get("ResultsNotFound"):
            self.log.info(
                "AWS Glue data quality ruleset evaluation run, results not found for %s",
                results["ResultsNotFound"],
            )

        for result in results["Results"]:
            rule_results = result["RuleResults"]

            total_failed_rules += len(
                list(
                    filter(
                        lambda result: result.get("Result") == "FAIL" or result.get("Result") == "ERROR",
                        rule_results,
                    )
                )
            )

            if show_results:
                self._log_results(result)

        self.log.info(
            "AWS Glue data quality ruleset evaluation run, total number of rules failed: %s",
            total_failed_rules,
        )

        if verify_result_status and total_failed_rules > 0:
            raise AirflowException(
                "AWS Glue data quality ruleset evaluation run failed for one or more rules"
            )

    def log_recommendation_results(self, run_id: str) -> None:
        """
        Print the outcome of recommendation run, recommendation run generates multiple rules against a data source (Glue table) in Data Quality Definition Language (DQDL) format.

        Rules = [
        IsComplete "NAME",
        ColumnLength "EMP_ID" between 1 and 12,
        IsUnique "EMP_ID",
        ColumnValues "INCOME" > 50000
        ]
        """
        result = self.conn.get_data_quality_rule_recommendation_run(RunId=run_id)

        if result.get("RecommendedRuleset"):
            self.log.info(
                "AWS Glue data quality recommended rules for DatabaseName: %s TableName: %s",
                result["DataSource"]["GlueTable"]["DatabaseName"],
                result["DataSource"]["GlueTable"]["TableName"],
            )
            self.log.info(result["RecommendedRuleset"])
        else:
            self.log.info("AWS Glue data quality, no recommended rules available for RunId: %s", run_id)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from datetime import timedelta
from typing import TYPE_CHECKING, Any

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.glue_catalog import GlueCatalogHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.glue import GlueCatalogPartitionTrigger
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields

if TYPE_CHECKING:
    from airflow.utils.context import Context


class GlueCatalogPartitionSensor(AwsBaseSensor[GlueCatalogHook]):
    """
    Waits for a partition to show up in AWS Glue Catalog.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:GlueCatalogPartitionSensor`

    :param table_name: The name of the table to wait for, supports the dot
        notation (my_database.my_table)
    :param expression: The partition clause to wait for. This is passed as
        is to the AWS Glue Catalog API's get_partitions function,
        and supports SQL like notation as in ``ds='2015-01-01'
        AND type='value'`` and comparison operators as in ``"ds>=2015-01-01"``.
        See https://docs.aws.amazon.com/glue/latest/dg/aws-glue-api-catalog-partitions.html
        #aws-glue-api-catalog-partitions-GetPartitions
    :param database_name: The name of the catalog database where the partitions reside.
    :param poke_interval: Time in seconds that the job should wait in
        between each tries
    :param deferrable: If true, then the sensor will wait asynchronously for the partition to
        show up in the AWS Glue Catalog.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = GlueCatalogHook

    template_fields: Sequence[str] = aws_template_fields(
        "database_name",
        "table_name",
        "expression",
    )
    ui_color = "#C5CAE9"

    def __init__(
        self,
        *,
        table_name: str,
        expression: str = "ds='{{ ds }}'",
        database_name: str = "default",
        poke_interval: int = 60 * 3,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.table_name = table_name
        self.expression = expression
        self.database_name = database_name
        self.poke_interval = poke_interval
        self.deferrable = deferrable

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=GlueCatalogPartitionTrigger(
                    database_name=self.database_name,
                    table_name=self.table_name,
                    expression=self.expression,
                    aws_conn_id=self.aws_conn_id,
                    region_name=self.region_name,
                    waiter_delay=int(self.poke_interval),
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="execute_complete",
                timeout=timedelta(seconds=self.timeout),
            )
        else:
            super().execute(context=context)

    def poke(self, context: Context):
        """Check for existence of the partition in the AWS Glue Catalog table."""
        if "." in self.table_name:
            self.database_name, self.table_name = self.table_name.split(".")
        self.log.info(
            "Poking for table %s. %s, expression %s", self.database_name, self.table_name, self.expression
        )

        return self.hook.check_for_partition(self.database_name, self.table_name, self.expression)

    def execute_complete(self, context: Context, event: dict | None = None) -> None:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException(f"Trigger error: event is {validated_event}")
        self.log.info("Partition exists in the Glue Catalog")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os

from airflow.providers.amazon.aws.utils.sagemaker_unified_studio import is_local_runner, workflows_env_key


def test_is_local_runner_false():
    assert not is_local_runner()


def test_is_local_runner_true():
    os.environ[workflows_env_key] = "Local"
    assert is_local_runner()


def test_is_local_runner_false_with_env_var():
    os.environ[workflows_env_key] = "False"
    assert not is_local_runner()


def test_is_local_runner_false_with_env_var_empty():
    os.environ[workflows_env_key] = ""
    assert not is_local_runner()


def test_is_local_runner_false_with_env_var_invalid():
    os.environ[workflows_env_key] = "random string"
    assert not is_local_runner()


def test_is_local_runner_false_with_string_int():
    os.environ[workflows_env_key] = "1"
    assert not is_local_runner()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import json
from datetime import datetime
from typing import Any

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.emr import (
    EmrAddStepsOperator,
    EmrCreateJobFlowOperator,
    EmrModifyClusterOperator,
    EmrTerminateJobFlowOperator,
)
from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.providers.amazon.aws.sensors.emr import EmrJobFlowSensor, EmrStepSensor
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_emr"
CONFIG_NAME = "EMR Runtime Role Security Configuration"
EXECUTION_ROLE_ARN_KEY = "EXECUTION_ROLE_ARN"
BUCKET_NAME_KEY = "BUCKET_NAME"

SECURITY_CONFIGURATION = {
    "AuthorizationConfiguration": {
        "IAMConfiguration": {
            "EnableApplicationScopedIAMRole": True,
        },
    },
    # Use IMDSv2 for greater security, see the following doc for more details:
    # https://docs.aws.amazon.com/emr/latest/ManagementGuide/emr-create-security-configuration.html
    "InstanceMetadataServiceConfiguration": {
        "MinimumInstanceMetadataServiceVersion": 2,
        "HttpPutResponseHopLimit": 2,
    },
}

# [START howto_operator_emr_steps_config]
SPARK_STEPS = [
    {
        "Name": "calculate_pi",
        "ActionOnFailure": "CONTINUE",
        "HadoopJarStep": {
            "Jar": "command-runner.jar",
            "Args": ["/usr/lib/spark/bin/run-example", "SparkPi", "10"],
        },
    }
]

JOB_FLOW_OVERRIDES: dict[str, Any] = {
    "Name": "PiCalc",
    "ReleaseLabel": "emr-7.1.0",
    "Applications": [{"Name": "Spark"}],
    "Instances": {
        "InstanceGroups": [
            {
                "Name": "Primary node",
                "Market": "ON_DEMAND",
                "InstanceRole": "MASTER",
                "InstanceType": "m5.xlarge",
                "InstanceCount": 1,
            },
        ],
        # If the EMR steps complete too quickly the cluster will be torn down before the other system test
        # tasks have a chance to run (such as the modify cluster step, the addition of more EMR steps, etc).
        # Set KeepJobFlowAliveWhenNoSteps to False to avoid the cluster from being torn down prematurely.
        "KeepJobFlowAliveWhenNoSteps": True,
        "TerminationProtected": False,
    },
    "Steps": SPARK_STEPS,
    "JobFlowRole": "EMR_EC2_DefaultRole",
    "ServiceRole": "EMR_DefaultRole",
}
# [END howto_operator_emr_steps_config]


@task
def configure_security_config(config_name: str):
    boto3.client("emr").create_security_configuration(
        Name=config_name,
        SecurityConfiguration=json.dumps(SECURITY_CONFIGURATION),
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_security_config(config_name: str):
    boto3.client("emr").delete_security_configuration(
        Name=config_name,
    )


@task
def get_step_id(step_ids: list):
    return step_ids[0]


sys_test_context_task = (
    SystemTestContextBuilder()
    .add_variable(EXECUTION_ROLE_ARN_KEY)
    .add_variable(BUCKET_NAME_KEY, optional=True)
    .build()
)

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule="@once",
    catchup=False,
    tags=["example"],
) as dag:
    test_context = sys_test_context_task()

    env_id = test_context[ENV_ID_KEY]
    config_name = f"{CONFIG_NAME}-{env_id}"
    execution_role_arn = test_context[EXECUTION_ROLE_ARN_KEY]
    s3_bucket = test_context[BUCKET_NAME_KEY] or f"{env_id}-emr-bucket"

    JOB_FLOW_OVERRIDES["LogUri"] = f"s3://{s3_bucket}/"
    JOB_FLOW_OVERRIDES["SecurityConfiguration"] = config_name

    create_security_configuration = configure_security_config(config_name)

    # [START howto_operator_emr_create_job_flow]
    create_job_flow = EmrCreateJobFlowOperator(
        task_id="create_job_flow",
        job_flow_overrides=JOB_FLOW_OVERRIDES,
    )
    # [END howto_operator_emr_create_job_flow]

    # [START howto_operator_emr_modify_cluster]
    modify_cluster = EmrModifyClusterOperator(
        task_id="modify_cluster", cluster_id=create_job_flow.output, step_concurrency_level=1
    )
    # [END howto_operator_emr_modify_cluster]

    # [START howto_operator_emr_add_steps]
    add_steps = EmrAddStepsOperator(
        task_id="add_steps",
        job_flow_id=create_job_flow.output,
        steps=SPARK_STEPS,
        execution_role_arn=execution_role_arn,
    )
    # [END howto_operator_emr_add_steps]
    add_steps.wait_for_completion = True

    # [START howto_sensor_emr_step]
    wait_for_step = EmrStepSensor(
        task_id="wait_for_step",
        job_flow_id=create_job_flow.output,
        step_id=get_step_id(add_steps.output),
    )
    # [END howto_sensor_emr_step]

    # [START howto_operator_emr_terminate_job_flow]
    remove_cluster = EmrTerminateJobFlowOperator(
        task_id="remove_cluster",
        job_flow_id=create_job_flow.output,
    )
    # [END howto_operator_emr_terminate_job_flow]
    remove_cluster.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_sensor_emr_job_flow]
    check_job_flow = EmrJobFlowSensor(task_id="check_job_flow", job_flow_id=create_job_flow.output)
    # [END howto_sensor_emr_job_flow]
    check_job_flow.poke_interval = 10

    delete_security_configuration = delete_security_config(config_name)

    # There are two options:
    # - Pass the bucket name as an argument to the system test. This bucket will then be used to store
    #   the EMR-related logs.
    # - The test itself creates and delete the S3 bucket needed for this test.
    create_s3_bucket: S3CreateBucketOperator | None = None
    delete_s3_bucket: S3DeleteBucketOperator | None = None
    if not test_context[BUCKET_NAME_KEY]:
        create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=s3_bucket)
        delete_s3_bucket = S3DeleteBucketOperator(
            task_id="delete_s3_bucket",
            bucket_name=s3_bucket,
            force_delete=True,
            trigger_rule=TriggerRule.ALL_DONE,
        )

    chain(
        *[
            task
            for task in [
                # TEST SETUP
                test_context,
                create_s3_bucket,
                create_security_configuration,
                # TEST BODY
                create_job_flow,
                modify_cluster,
                add_steps,
                wait_for_step,
                # TEST TEARDOWN
                remove_cluster,
                check_job_flow,
                delete_security_configuration,
                delete_s3_bucket,
            ]
            if task is not None
        ]
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from airflow.providers.amazon.aws.triggers.batch import BatchJobTrigger


class TestBatchJobTrigger:
    def test_serialization(self):
        job_id = "test_job_id"
        aws_conn_id = "aws_default"
        region_name = "us-west-2"

        trigger = BatchJobTrigger(
            job_id=job_id,
            aws_conn_id=aws_conn_id,
            region_name=region_name,
        )

        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.amazon.aws.triggers.batch.BatchJobTrigger"
        assert kwargs == {
            "job_id": "test_job_id",
            "waiter_delay": 5,
            "waiter_max_attempts": 720,
            "aws_conn_id": "aws_default",
            "region_name": "us-west-2",
        }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import boto3
import botocore
import pytest

from airflow.providers.amazon.aws.hooks.glue import GlueDataQualityHook
from airflow.providers.amazon.aws.sensors.glue import (
    GlueDataQualityRuleRecommendationRunSensor,
    GlueDataQualityRuleSetEvaluationRunSensor,
)


class TestGlueDataQualityCustomWaiters:
    def test_evaluation_run_waiters(self):
        assert "data_quality_ruleset_evaluation_run_complete" in GlueDataQualityHook().list_waiters()

    def test_recommendation_run_waiters(self):
        assert "data_quality_rule_recommendation_run_complete" in GlueDataQualityHook().list_waiters()


class TestGlueDataQualityCustomWaitersBase:
    @pytest.fixture(autouse=True)
    def mock_conn(self, monkeypatch):
        self.client = boto3.client("glue")
        monkeypatch.setattr(GlueDataQualityHook, "conn", self.client)


class TestGlueDataQualityRuleSetEvaluationRunCompleteWaiter(TestGlueDataQualityCustomWaitersBase):
    WAITER_NAME = "data_quality_ruleset_evaluation_run_complete"

    @pytest.fixture
    def mock_get_job(self):
        with mock.patch.object(self.client, "get_data_quality_ruleset_evaluation_run") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", GlueDataQualityRuleSetEvaluationRunSensor.SUCCESS_STATES)
    def test_data_quality_ruleset_evaluation_run_complete(self, state, mock_get_job):
        mock_get_job.return_value = {"Status": state}

        GlueDataQualityHook().get_waiter(self.WAITER_NAME).wait(RunId="run_id")

    @pytest.mark.parametrize("state", GlueDataQualityRuleSetEvaluationRunSensor.FAILURE_STATES)
    def test_data_quality_ruleset_evaluation_run_failed(self, state, mock_get_job):
        mock_get_job.return_value = {"Status": state}

        with pytest.raises(botocore.exceptions.WaiterError):
            GlueDataQualityHook().get_waiter(self.WAITER_NAME).wait(RunId="run_id")

    def test_data_quality_ruleset_evaluation_run_wait(self, mock_get_job):
        wait = {"Status": "RUNNING"}
        success = {"Status": "SUCCEEDED"}
        mock_get_job.side_effect = [wait, wait, success]

        GlueDataQualityHook().get_waiter(self.WAITER_NAME).wait(
            RunIc="run_id", WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )


class TestGlueDataQualityRuleRecommendationRunCompleteWaiter(TestGlueDataQualityCustomWaitersBase):
    WAITER_NAME = "data_quality_rule_recommendation_run_complete"

    @pytest.fixture
    def mock_get_job(self):
        with mock.patch.object(self.client, "get_data_quality_rule_recommendation_run") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", GlueDataQualityRuleRecommendationRunSensor.SUCCESS_STATES)
    def test_data_quality_rule_recommendation_run_complete(self, state, mock_get_job):
        mock_get_job.return_value = {"Status": state}

        GlueDataQualityHook().get_waiter(self.WAITER_NAME).wait(RunId="run_id")

    @pytest.mark.parametrize("state", GlueDataQualityRuleRecommendationRunSensor.FAILURE_STATES)
    def test_data_quality_rule_recommendation_run_failed(self, state, mock_get_job):
        mock_get_job.return_value = {"Status": state}

        with pytest.raises(botocore.exceptions.WaiterError):
            GlueDataQualityHook().get_waiter(self.WAITER_NAME).wait(RunId="run_id")

    def test_data_quality_rule_recommendation_run_wait(self, mock_get_job):
        wait = {"Status": "RUNNING"}
        success = {"Status": "SUCCEEDED"}
        mock_get_job.side_effect = [wait, wait, success]

        GlueDataQualityHook().get_waiter(self.WAITER_NAME).wait(
            RunIc="run_id", WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.jobs.triggerer_job_runner import TriggerRunner
from airflow.models import Trigger
from airflow.providers.amazon.aws.triggers.athena import AthenaTrigger
from airflow.providers.amazon.aws.triggers.batch import BatchCreateComputeEnvironmentTrigger, BatchJobTrigger
from airflow.providers.amazon.aws.triggers.ecs import ClusterActiveTrigger, ClusterInactiveTrigger
from airflow.providers.amazon.aws.triggers.eks import (
    EksCreateClusterTrigger,
    EksCreateFargateProfileTrigger,
    EksCreateNodegroupTrigger,
    EksDeleteClusterTrigger,
    EksDeleteFargateProfileTrigger,
    EksDeleteNodegroupTrigger,
)
from airflow.providers.amazon.aws.triggers.emr import (
    EmrAddStepsTrigger,
    EmrContainerTrigger,
    EmrCreateJobFlowTrigger,
    EmrServerlessCancelJobsTrigger,
    EmrServerlessCreateApplicationTrigger,
    EmrServerlessDeleteApplicationTrigger,
    EmrServerlessStartApplicationTrigger,
    EmrServerlessStartJobTrigger,
    EmrServerlessStopApplicationTrigger,
    EmrStepSensorTrigger,
    EmrTerminateJobFlowTrigger,
)
from airflow.providers.amazon.aws.triggers.glue import GlueCatalogPartitionTrigger
from airflow.providers.amazon.aws.triggers.glue_crawler import GlueCrawlerCompleteTrigger
from airflow.providers.amazon.aws.triggers.lambda_function import LambdaCreateFunctionCompleteTrigger
from airflow.providers.amazon.aws.triggers.rds import (
    RdsDbAvailableTrigger,
    RdsDbDeletedTrigger,
    RdsDbStoppedTrigger,
)
from airflow.providers.amazon.aws.triggers.redshift_cluster import (
    RedshiftCreateClusterSnapshotTrigger,
    RedshiftCreateClusterTrigger,
    RedshiftDeleteClusterTrigger,
    RedshiftPauseClusterTrigger,
    RedshiftResumeClusterTrigger,
)
from airflow.providers.amazon.aws.triggers.sqs import SqsSensorTrigger
from airflow.providers.amazon.aws.triggers.step_function import StepFunctionsExecutionCompleteTrigger
from airflow.providers.amazon.aws.utils.rds import RdsDbType
from airflow.serialization.serialized_objects import BaseSerialization

pytestmark = pytest.mark.db_test

BATCH_JOB_ID = "job_id"

TEST_CLUSTER_IDENTIFIER = "test-cluster"
TEST_FARGATE_PROFILE_NAME = "test-fargate-profile"
TEST_NODEGROUP_NAME = "test-nodegroup"

TEST_JOB_FLOW_ID = "test-job-flow-id"
VIRTUAL_CLUSTER_ID = "vzwemreks"
JOB_ID = "job-1234"
TARGET_STATE = ["TERMINATED"]
STEP_ID = "s-1234"

TEST_APPLICATION_ID = "test-application-id"
TEST_JOB_ID = "test-job-id"

TEST_FUNCTION_NAME = "test-function-name"

TEST_DB_INSTANCE_IDENTIFIER = "test-db-instance-identifier"
TEST_RESPONSE = {
    "DBInstance": {
        "DBInstanceIdentifier": "test-db-instance-identifier",
        "DBInstanceStatus": "test-db-instance-status",
    }
}

TEST_SQS_QUEUE = "test-sqs-queue"
TEST_MAX_MESSAGES = 1
TEST_NUM_BATCHES = 1
TEST_WAIT_TIME_SECONDS = 1
TEST_VISIBILITY_TIMEOUT = 1
TEST_MESSAGE_FILTERING_MATCH_VALUES = "test"
TEST_MESSAGE_FILTERING_CONFIG = "test-message-filtering-config"
TEST_DELETE_MESSAGE_ON_RECEPTION = False

TEST_ARN = "test-aws-arn"

WAITER_DELAY = 5
MAX_ATTEMPTS = 5
AWS_CONN_ID = "aws_batch_job_conn"
AWS_REGION = "us-east-2"


def gen_test_name(trigger):
    """Gives to tests the name of the class being tested."""
    return trigger.__class__.__name__


class TestTriggersSerialization:
    @pytest.mark.parametrize(
        "trigger",
        [
            AthenaTrigger("query_id", 1, 5, "aws connection"),
            BatchJobTrigger(
                job_id=BATCH_JOB_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
                region_name=AWS_REGION,
            ),
            BatchCreateComputeEnvironmentTrigger(
                compute_env_arn="my_arn",
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
                region_name=AWS_REGION,
            ),
            ClusterActiveTrigger(
                cluster_arn="my_arn",
                aws_conn_id="my_conn",
                waiter_delay=1,
                waiter_max_attempts=2,
                region_name="my_region",
            ),
            ClusterInactiveTrigger(
                cluster_arn="my_arn",
                aws_conn_id="my_conn",
                waiter_delay=1,
                waiter_max_attempts=2,
                region_name="my_region",
            ),
            EksCreateFargateProfileTrigger(
                cluster_name=TEST_CLUSTER_IDENTIFIER,
                fargate_profile_name=TEST_FARGATE_PROFILE_NAME,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EksDeleteFargateProfileTrigger(
                cluster_name=TEST_CLUSTER_IDENTIFIER,
                fargate_profile_name=TEST_FARGATE_PROFILE_NAME,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EksCreateNodegroupTrigger(
                cluster_name=TEST_CLUSTER_IDENTIFIER,
                nodegroup_name=TEST_NODEGROUP_NAME,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                region_name=AWS_REGION,
            ),
            EksDeleteNodegroupTrigger(
                cluster_name=TEST_CLUSTER_IDENTIFIER,
                nodegroup_name=TEST_NODEGROUP_NAME,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                region_name=AWS_REGION,
            ),
            EksCreateClusterTrigger(
                cluster_name=TEST_CLUSTER_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=WAITER_DELAY,
                aws_conn_id=AWS_CONN_ID,
                region_name=AWS_REGION,
            ),
            EksDeleteClusterTrigger(
                cluster_name=TEST_CLUSTER_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=WAITER_DELAY,
                aws_conn_id=AWS_CONN_ID,
                region_name=AWS_REGION,
                force_delete_compute=True,
            ),
            EmrAddStepsTrigger(
                job_flow_id=TEST_JOB_FLOW_ID,
                step_ids=["my_step1", "my_step2"],
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EmrCreateJobFlowTrigger(
                job_flow_id=TEST_JOB_FLOW_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EmrTerminateJobFlowTrigger(
                job_flow_id=TEST_JOB_FLOW_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EmrContainerTrigger(
                virtual_cluster_id=VIRTUAL_CLUSTER_ID,
                job_id=JOB_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
            ),
            EmrStepSensorTrigger(
                job_flow_id=TEST_JOB_FLOW_ID,
                step_id=STEP_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
            ),
            EmrServerlessCreateApplicationTrigger(
                application_id=TEST_APPLICATION_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EmrServerlessStartApplicationTrigger(
                application_id=TEST_APPLICATION_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EmrServerlessStopApplicationTrigger(
                application_id=TEST_APPLICATION_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EmrServerlessDeleteApplicationTrigger(
                application_id=TEST_APPLICATION_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EmrServerlessCancelJobsTrigger(
                application_id=TEST_APPLICATION_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            EmrServerlessStartJobTrigger(
                application_id=TEST_APPLICATION_ID,
                job_id=TEST_JOB_ID,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            GlueCrawlerCompleteTrigger(crawler_name="my_crawler", waiter_delay=2, aws_conn_id="my_conn_id"),
            GlueCatalogPartitionTrigger(
                database_name="my_database",
                table_name="my_table",
                expression="my_expression",
                aws_conn_id="my_conn_id",
            ),
            LambdaCreateFunctionCompleteTrigger(
                function_name=TEST_FUNCTION_NAME,
                function_arn=TEST_ARN,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
            ),
            RedshiftCreateClusterTrigger(
                cluster_identifier=TEST_CLUSTER_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
            ),
            RedshiftPauseClusterTrigger(
                cluster_identifier=TEST_CLUSTER_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
            ),
            RedshiftCreateClusterSnapshotTrigger(
                cluster_identifier=TEST_CLUSTER_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
            ),
            RedshiftResumeClusterTrigger(
                cluster_identifier=TEST_CLUSTER_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
            ),
            RedshiftDeleteClusterTrigger(
                cluster_identifier=TEST_CLUSTER_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
            ),
            RdsDbAvailableTrigger(
                db_identifier=TEST_DB_INSTANCE_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
                region_name=AWS_REGION,
                response=TEST_RESPONSE,
                db_type=RdsDbType.INSTANCE,
            ),
            RdsDbDeletedTrigger(
                db_identifier=TEST_DB_INSTANCE_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
                region_name=AWS_REGION,
                response=TEST_RESPONSE,
                db_type=RdsDbType.INSTANCE,
            ),
            RdsDbStoppedTrigger(
                db_identifier=TEST_DB_INSTANCE_IDENTIFIER,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                aws_conn_id=AWS_CONN_ID,
                region_name=AWS_REGION,
                response=TEST_RESPONSE,
                db_type=RdsDbType.INSTANCE,
            ),
            SqsSensorTrigger(
                sqs_queue=TEST_SQS_QUEUE,
                aws_conn_id=AWS_CONN_ID,
                max_messages=TEST_MAX_MESSAGES,
                num_batches=TEST_NUM_BATCHES,
                wait_time_seconds=TEST_WAIT_TIME_SECONDS,
                visibility_timeout=TEST_VISIBILITY_TIMEOUT,
                message_filtering="literal",
                message_filtering_match_values=TEST_MESSAGE_FILTERING_MATCH_VALUES,
                message_filtering_config=TEST_MESSAGE_FILTERING_CONFIG,
                delete_message_on_reception=TEST_DELETE_MESSAGE_ON_RECEPTION,
                waiter_delay=WAITER_DELAY,
            ),
            StepFunctionsExecutionCompleteTrigger(
                execution_arn=TEST_ARN,
                aws_conn_id=AWS_CONN_ID,
                waiter_delay=WAITER_DELAY,
                waiter_max_attempts=MAX_ATTEMPTS,
                region_name=AWS_REGION,
            ),
        ],
        ids=gen_test_name,
    )
    def test_serialize_recreate(self, trigger):
        # generate the DB object from the trigger
        trigger_db: Trigger = Trigger.from_object(trigger)

        # serialize/deserialize using the same method that is used when inserting in DB
        json_params = BaseSerialization.serialize(trigger_db.kwargs)
        retrieved_params = BaseSerialization.deserialize(json_params)

        # recreate a new trigger object from the data we would have in DB
        clazz = TriggerRunner().get_trigger_by_classpath(trigger_db.classpath)
        # noinspection PyArgumentList
        instance = clazz(**retrieved_params)

        # recreate a DB column object from the new trigger so that we can easily compare attributes
        trigger_db_2: Trigger = Trigger.from_object(instance)

        assert trigger_db.classpath == trigger_db_2.classpath
        assert trigger_db.kwargs == trigger_db_2.kwargs



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.comprehend import ComprehendHook


class TestComprehendHook:
    @pytest.mark.parametrize(
        "test_hook, service_name",
        [pytest.param(ComprehendHook(), "comprehend", id="comprehend")],
    )
    def test_comprehend_hook(self, test_hook, service_name):
        comprehend_hook = ComprehendHook()
        assert comprehend_hook.conn is not None

    @mock.patch.object(ComprehendHook, "conn")
    def test_validate_document_classifier_training_status_for_success(self, mock_conn):
        mock_conn.describe_document_classifier.return_value = {
            "DocumentClassifierProperties": {
                "DocumentClassifierArn": "arn:aws:comprehend:us-east-1:123456789012:document-classifier/docs-clasf/version/v1",
                "LanguageCode": "en",
                "Status": "TRAINED",
                "Message": "",
                "SubmitTime": "2024-06-08T21:41:36.867Z",
                "EndTime": "2024-06-08T21:52:02.249Z",
                "TrainingStartTime": "2024-06-08T21:44:39.596Z",
                "TrainingEndTime": "2024-06-08T21:51:00.021Z",
                "InputDataConfig": {
                    "DataFormat": "COMPREHEND_CSV",
                    "S3Uri": "s3://test/native-doc.csv",
                    "DocumentType": "SEMI_STRUCTURED_DOCUMENT",
                    "Documents": {"S3Uri": "s3://test/input-docs/"},
                    "DocumentReaderConfig": {
                        "DocumentReadAction": "TEXTRACT_DETECT_DOCUMENT_TEXT",
                        "DocumentReadMode": "SERVICE_DEFAULT",
                    },
                },
                "OutputDataConfig": {
                    "S3Uri": "s3://test/training_output/183167903796-CLR-b56fd4cf0b5bcc11c9409dfb431cd585/output/output.tar.gz"
                },
                "ClassifierMetadata": {
                    "NumberOfLabels": 2,
                    "NumberOfTrainedDocuments": 16,
                    "NumberOfTestDocuments": 2,
                    "EvaluationMetrics": {
                        "Accuracy": 1,
                        "Precision": 1,
                        "Recall": 1,
                        "F1Score": 1,
                        "MicroPrecision": 1,
                        "MicroRecall": 1,
                        "MicroF1Score": 1,
                        "HammingLoss": 0,
                    },
                },
                "DataAccessRoleArn": "arn:aws:iam::123456789012:role/ComprehendExecutionRole",
                "Mode": "MULTI_CLASS",
                "VersionName": "v1",
            }
        }
        classifier_arn = "arn:aws:comprehend:us-east-1:123456789012:document-classifier/docs-clasf/version/v1"
        ComprehendHook().validate_document_classifier_training_status(document_classifier_arn=classifier_arn)

    @mock.patch.object(ComprehendHook, "conn")
    def test_validate_document_classifier_training_status_for_warning(self, mock_conn):
        mock_conn.describe_document_classifier.return_value = {
            "DocumentClassifierProperties": {
                "DocumentClassifierArn": "arn:aws:comprehend:us-east-1:123456789012:document-classifier/docs-clasf/version/v1",
                "LanguageCode": "en",
                "Status": "TRAINED_WITH_WARNING",
                "Message": "Unable to parse some documents. See details in the output S3 location. These documents have been skipped.",
                "SubmitTime": "2024-06-08T21:41:36.867Z",
                "EndTime": "2024-06-08T21:52:02.249Z",
                "TrainingStartTime": "2024-06-08T21:44:39.596Z",
                "TrainingEndTime": "2024-06-08T21:51:00.021Z",
                "InputDataConfig": {
                    "DataFormat": "COMPREHEND_CSV",
                    "S3Uri": "s3://test/native-doc.csv",
                    "DocumentType": "SEMI_STRUCTURED_DOCUMENT",
                    "Documents": {"S3Uri": "s3://test/input-docs/"},
                    "DocumentReaderConfig": {
                        "DocumentReadAction": "TEXTRACT_DETECT_DOCUMENT_TEXT",
                        "DocumentReadMode": "SERVICE_DEFAULT",
                    },
                },
                "OutputDataConfig": {
                    "S3Uri": "s3://test/training_output/183167903796-CLR-b56fd4cf0b5bcc11c9409dfb431cd585/output/output.tar.gz"
                },
                "ClassifierMetadata": {
                    "NumberOfLabels": 2,
                    "NumberOfTrainedDocuments": 16,
                    "NumberOfTestDocuments": 2,
                    "EvaluationMetrics": {
                        "Accuracy": 1,
                        "Precision": 1,
                        "Recall": 1,
                        "F1Score": 1,
                        "MicroPrecision": 1,
                        "MicroRecall": 1,
                        "MicroF1Score": 1,
                        "HammingLoss": 0,
                    },
                },
                "DataAccessRoleArn": "arn:aws:iam::123456789012:role/ComprehendExecutionRole",
                "Mode": "MULTI_CLASS",
                "VersionName": "v1",
            }
        }
        classifier_arn = "arn:aws:comprehend:us-east-1:123456789012:document-classifier/docs-clasf/version/v1"
        with pytest.raises(
            AirflowException, match="Warnings in AWS Comprehend document classifier training."
        ):
            ComprehendHook().validate_document_classifier_training_status(
                document_classifier_arn=classifier_arn, fail_on_warnings=True
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import functools
import inspect
import json
import logging
import os
import time
from pathlib import Path
from typing import TYPE_CHECKING
from uuid import uuid4

import boto3
from botocore.exceptions import ClientError, NoCredentialsError

from airflow.decorators import task
from airflow.providers.amazon.aws.hooks.ssm import SsmHook
from airflow.providers.amazon.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.utils.state import DagRunState, State
from airflow.utils.trigger_rule import TriggerRule

if TYPE_CHECKING:
    from botocore.client import BaseClient

ENV_ID_ENVIRON_KEY: str = "SYSTEM_TESTS_ENV_ID"
ENV_ID_KEY: str = "ENV_ID"
DEFAULT_ENV_ID_PREFIX: str = "env"
DEFAULT_ENV_ID_LEN: int = 8
DEFAULT_ENV_ID: str = f"{DEFAULT_ENV_ID_PREFIX}{uuid4()!s:.{DEFAULT_ENV_ID_LEN}}"
PURGE_LOGS_INTERVAL_PERIOD = 5

# All test file names will contain one of these strings.
TEST_FILE_IDENTIFIERS: list[str] = ["example_", "test_"]

INVALID_ENV_ID_MSG: str = (
    "To maximize compatibility, the SYSTEM_TESTS_ENV_ID must be an alphanumeric string "
    "which starts with a letter. Please see `providers/amazon/tests/system/amazon/README.md`."
)
LOWERCASE_ENV_ID_MSG: str = (
    "The provided Environment ID contains uppercase letters and "
    "will be converted to lowercase for the AWS System Tests."
)
NO_VALUE_MSG: str = "No Value Found: Variable {key} could not be found, and no default value was provided."

log = logging.getLogger(__name__)


def _get_test_name() -> str:
    """
    Extracts the module name from the test module.

    :return: The name of the test module that called the helper method.
    """
    # The exact layer of the stack will depend on if this is called directly
    # or from another helper, but the test will always contain the identifier.
    test_filename: str = next(
        frame.filename
        for frame in inspect.stack()
        if any(identifier in frame.filename for identifier in TEST_FILE_IDENTIFIERS)
    )
    return Path(test_filename).stem


def _validate_env_id(env_id: str) -> str:
    """
    Verifies that a prospective Environment ID value fits requirements.
    An Environment ID for an AWS System test must be a lowercase alphanumeric
    string which starts with a letter.

    :param env_id: An Environment ID to validate.
    :return: A validated string cast to lowercase.
    """
    if any(char.isupper() for char in str(env_id)):
        print(LOWERCASE_ENV_ID_MSG)
    if not env_id.isalnum() or not env_id[0].isalpha():
        raise ValueError(INVALID_ENV_ID_MSG)

    return env_id.lower()


@functools.cache
def _fetch_from_ssm(key: str, test_name: str | None = None) -> str:
    """
    Test values are stored in the SSM Value as a JSON-encoded dict of key/value pairs.

    :param key: The key to search for within the returned Parameter Value.
    :return: The value of the provided key from SSM
    """
    _test_name: str = test_name or _get_test_name()
    hook = SsmHook(aws_conn_id=None)
    value: str = ""

    try:
        value = json.loads(hook.get_parameter_value(_test_name))[key]
    # Since a default value after the SSM check is allowed, these exceptions should not stop execution.
    except NoCredentialsError as e:
        log.info("No boto credentials found: %s", e)
    except ClientError as e:
        log.info("Client error when connecting to SSM: %s", e)
    except hook.conn.exceptions.ParameterNotFound as e:
        log.info("SSM does not contain any parameter for this test: %s", e)
    except KeyError as e:
        log.info(
            "SSM contains one parameter for this test, but not the requested value: %s",
            e,
        )
    return value


class Variable:
    """
    Stores metadata about a variable to be fetched for AWS System Tests.

    :param name: The name of the variable to be fetched.
    :param to_split: If True, the input is a string-formatted List and needs to be split. Defaults to False.
    :param delimiter: If to_split is true, this will be used to split the string. Defaults to ','.
    :param test_name: The name of the system test that the variable is associated with.
    """

    def __init__(
        self,
        name: str,
        to_split: bool = False,
        delimiter: str | None = None,
        test_name: str | None = None,
        optional: bool = False,
    ):
        self.name = name
        self.test_name = test_name
        self.to_split = to_split
        if to_split:
            self.delimiter = delimiter or ","
        elif delimiter:
            raise ValueError(f"Variable {name} has a delimiter but split_string is set to False.")

        self.optional = optional

    def get_value(self):
        if hasattr(self, "default_value"):
            return self._format_value(
                fetch_variable(
                    key=self.name,
                    default_value=self.default_value,
                    test_name=self.test_name,
                    optional=self.optional,
                )
            )

        return self._format_value(
            fetch_variable(key=self.name, test_name=self.test_name, optional=self.optional)
        )

    def set_default(self, default):
        # Since 'None' is a potentially valid "default" value, we are only creating this
        # field when a default is provided, and in get_value we check if the field exists.
        self.default_value = default

    def _format_value(self, value):
        if self.to_split:
            if not isinstance(value, str):
                raise TypeError(f"{self.name} is type {type(value)} and can not be split as requested.")
            return value.split(self.delimiter)
        return value


class SystemTestContextBuilder:
    """
    This builder class ultimately constructs a TaskFlow task which is run at
    runtime (task execution time). This task generates and stores the test ENV_ID as well
    as any external resources requested (e.g.g IAM Roles, VPC, etc)
    """

    def __init__(self):
        self.variables = set()
        self.test_name = _get_test_name()

    def add_variable(
        self,
        variable_name: str,
        split_string: bool = False,
        delimiter: str | None = None,
        optional: bool = False,
        **kwargs,
    ):
        """Register a variable to fetch from environment or cloud parameter store"""
        if variable_name in [variable.name for variable in self.variables]:
            raise ValueError(f"Variable name {variable_name} already exists in the fetched variables list.")

        new_variable = Variable(
            name=variable_name,
            to_split=split_string,
            delimiter=delimiter,
            test_name=self.test_name,
            optional=optional,
        )

        # default_value is accepted via kwargs so that it is completely optional and no
        # default value needs to be provided in the method stub.  For example, if we
        # set it to None, we would have no way to know if that was our default or the
        # legitimate default value that the caller wishes to pass through.
        if "default_value" in kwargs:
            new_variable.set_default(kwargs["default_value"])

        self.variables.add(new_variable)

        return self  # Builder recipe; returning self allows chaining

    def build(self):
        """Build and return a TaskFlow task which will create an env_id and
        fetch requested variables. Storing everything in xcom for downstream
        tasks to use."""

        @task
        def variable_fetcher(ti=None):
            ti.xcom_push(ENV_ID_KEY, set_env_id())
            for variable in self.variables:
                ti.xcom_push(variable.name, variable.get_value())

        return variable_fetcher


def fetch_variable(
    key: str,
    default_value: str | None = None,
    test_name: str | None = None,
    optional: bool = False,
) -> str | None:
    """
    Given a Parameter name: first check for an existing Environment Variable,
    then check SSM for a value. If neither are available, fall back on the
    optional default value.

    :param key: The name of the Parameter to fetch a value for.
    :param default_value: The default value to use if no value can be found.
    :param test_name: The system test name.
    :param optional: Whether the variable is optional. If True, does not raise `ValueError` if variable
        does not exist
    :return: The value of the parameter.
    """

    value: str | None = os.getenv(key, _fetch_from_ssm(key, test_name)) or default_value
    if not optional and not value:
        raise ValueError(NO_VALUE_MSG.format(key=key))
    return value


def set_env_id() -> str:
    """
    Retrieves or generates an Environment ID, validate that it is suitable,
    export it as an Environment Variable, and return it.

    If an Environment ID has already been generated, use that.
    Otherwise, try to fetch it and export it as an Environment Variable.
    If there is not one available to fetch then generate one and export it as an Environment Variable.

    :return: A valid System Test Environment ID.
    """
    env_id: str = str(fetch_variable(ENV_ID_ENVIRON_KEY, DEFAULT_ENV_ID))
    env_id = _validate_env_id(env_id)

    os.environ[ENV_ID_ENVIRON_KEY] = env_id
    return env_id


def all_tasks_passed(ti) -> bool:
    if AIRFLOW_V_3_0_PLUS:
        # If the test is being run with task SDK, ti is an instance of RuntimeTaskInstance
        # This is the case when executed with an executor
        from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance

        if isinstance(ti, RuntimeTaskInstance):
            return RuntimeTaskInstance.get_dagrun_state(ti.dag_id, ti.run_id) != DagRunState.FAILED

    task_runs = ti.get_dagrun().get_task_instances()
    return all([_task.state != State.FAILED for _task in task_runs])


@task(trigger_rule=TriggerRule.ALL_DONE)
def prune_logs(
    logs: list[tuple[str, str | None]],
    force_delete: bool = False,
    retry: bool = False,
    retry_times: int = 3,
    delete_log_groups: bool = True,
    ti=None,
):
    """
    If all tasks in this dagrun have succeeded, then delete the associated logs.
    Otherwise, append the logs with a retention policy.  This allows the logs
    to be used for troubleshooting but assures they won't build up indefinitely.

    :param logs: A list of log_group/stream_prefix tuples to delete.
    :param force_delete: Whether to check log streams within the log group before
        removal. If True, removes the log group and all its log streams inside it.
    :param retry: Whether to retry if the log group/stream was not found. In some
        cases, the log group/stream is created seconds after the main resource has
        been created. By default, it retries for 3 times with a 5s waiting period.
    :param retry_times: Number of retries.
    :param delete_log_groups: Whether to delete the log groups if they are empty.
        Overridden by force_delete.
    :param ti: Used to check the status of the tasks. This gets pulled from the
        DAG's context and does not need to be passed manually.
    """
    if all_tasks_passed(ti):
        _purge_logs(logs, force_delete, retry, retry_times, delete_log_groups)
    else:
        client: BaseClient = boto3.client("logs")
        for group, _ in logs:
            client.put_retention_policy(logGroupName=group, retentionInDays=30)


def _purge_logs(
    test_logs: list[tuple[str, str | None]],
    force_delete: bool = False,
    retry: bool = False,
    retry_times: int = 3,
    delete_log_groups: bool = True,
) -> None:
    """
    Accepts a tuple in the format: ('log group name', 'log stream prefix').
    For each log group, it will delete any log streams matching the provided
    prefix then if the log group is empty, delete the group. If the group
    is not empty that indicates there are logs not generated by the test and
    those are left intact. If `check_log_streams` is True, it will simply delete the log group regardless
    of log streams within that log group.

    :param test_logs: A list of log_group/stream_prefix tuples to delete.
    :param force_delete: Whether to check log streams within the log group before removal. If True,
        removes the log group and all its log streams inside it
    :param retry: Whether to retry if the log group/stream was not found. In some cases, the log group/stream
        is created seconds after the main resource has been created. By default, it retries for 3 times
        with a 5s waiting period
    :param retry_times: Number of retries
    :param delete_log_groups: Whether to delete the log groups if they are empty.
        Overridden by force_delete.
    """
    client: BaseClient = boto3.client("logs")

    for group, prefix in test_logs:
        try:
            if prefix:
                log_streams = client.describe_log_streams(
                    logGroupName=group,
                    logStreamNamePrefix=prefix,
                )["logStreams"]

                for stream_name in [stream["logStreamName"] for stream in log_streams]:
                    client.delete_log_stream(logGroupName=group, logStreamName=stream_name)

            if force_delete or (
                delete_log_groups and not client.describe_log_streams(logGroupName=group)["logStreams"]
            ):
                client.delete_log_group(logGroupName=group)
        except ClientError as e:
            if not retry or retry_times == 0 or e.response["Error"]["Code"] != "ResourceNotFoundException":
                raise e

            time.sleep(PURGE_LOGS_INTERVAL_PERIOD)
            _purge_logs(
                test_logs=test_logs,
                force_delete=force_delete,
                retry=retry,
                retry_times=retry_times - 1,
            )


@task
def split_string(string):
    return string.split(",")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json

import boto3
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.eks import EksHook

from unit.amazon.aws.waiters.test_custom_waiters import assert_all_match


class TestCustomEKSServiceWaiters:
    def test_service_waiters(self):
        hook = EksHook()
        with open(hook.waiter_path) as config_file:
            expected_waiters = json.load(config_file)["waiters"]

        for waiter in list(expected_waiters.keys()):
            assert waiter in hook.list_waiters()
            assert waiter in hook._list_custom_waiters()

    @mock_aws
    def test_existing_waiter_inherited(self):
        """
        AwsBaseHook::get_waiter will first check if there is a custom waiter with the
        provided name and pass that through is it exists, otherwise it will check the
        custom waiters for the given service.  This test checks to make sure that the
        waiter is the same whichever way you get it and no modifications are made.
        """
        hook_waiter = EksHook().get_waiter("cluster_active")
        client_waiter = EksHook().conn.get_waiter("cluster_active")
        boto_waiter = boto3.client("eks").get_waiter("cluster_active")

        assert_all_match(hook_waiter.name, client_waiter.name, boto_waiter.name)
        assert_all_match(len(hook_waiter.__dict__), len(client_waiter.__dict__), len(boto_waiter.__dict__))
        for attr in hook_waiter.__dict__:
            # Not all attributes in a Waiter are directly comparable
            # so the best we can do it make sure the same attrs exist.
            assert hasattr(boto_waiter, attr)
            assert hasattr(client_waiter, attr)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.eventbridge import EventBridgeHook
from airflow.providers.amazon.aws.operators.eventbridge import (
    EventBridgeDisableRuleOperator,
    EventBridgeEnableRuleOperator,
    EventBridgePutEventsOperator,
    EventBridgePutRuleOperator,
)

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

if TYPE_CHECKING:
    from unittest.mock import MagicMock

ENTRIES = [{"Detail": "test-detail", "Source": "test-source", "DetailType": "test-detail-type"}]
FAILED_ENTRIES_RESPONSE = [{"ErrorCode": "test_code"}, {"ErrorCode": "test_code"}]
EVENT_PATTERN = '{"source": ["aws.s3"]}'
RULE_NAME = "match_s3_events"


class TestEventBridgePutEventsOperator:
    def test_init(self):
        op = EventBridgePutEventsOperator(
            task_id="put_events_job",
            entries=ENTRIES,
            aws_conn_id="fake-conn-id",
            region_name="eu-central-1",
            verify="/spam/egg.pem",
            botocore_config={"read_timeout": 42},
        )
        assert op.entries == ENTRIES
        assert op.hook.client_type == "events"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "eu-central-1"
        assert op.hook._verify == "/spam/egg.pem"
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = EventBridgePutEventsOperator(task_id="put_events_job", entries=ENTRIES)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(EventBridgeHook, "conn")
    def test_execute(self, mock_conn: MagicMock):
        hook_response = {"FailedEntryCount": 0, "Entries": [{"EventId": "foobar"}]}

        mock_conn.put_events.return_value = hook_response

        operator = EventBridgePutEventsOperator(
            task_id="put_events_job",
            entries=ENTRIES,
        )

        result = operator.execute(context={})

        assert result == ["foobar"]

    @mock.patch.object(EventBridgeHook, "conn")
    def test_failed_to_send(self, mock_conn: MagicMock):
        hook_response = {
            "FailedEntryCount": 1,
            "Entries": FAILED_ENTRIES_RESPONSE,
        }

        mock_conn.put_events.return_value = hook_response

        operator = EventBridgePutEventsOperator(
            task_id="failed_put_events_job",
            entries=ENTRIES,
        )

        with pytest.raises(AirflowException):
            operator.execute(context={})

    def test_template_fields(self):
        operator = EventBridgePutEventsOperator(
            task_id="failed_put_events_job",
            entries=ENTRIES,
        )
        validate_template_fields(operator)


class TestEventBridgePutRuleOperator:
    def test_init(self):
        op = EventBridgePutRuleOperator(
            task_id="events_put_rule_job",
            name=RULE_NAME,
            event_pattern=EVENT_PATTERN,
            aws_conn_id="fake-conn-id",
            region_name="eu-west-1",
            verify="/spam/egg.pem",
            botocore_config={"read_timeout": 42},
        )
        assert op.event_pattern == EVENT_PATTERN
        assert op.hook.client_type == "events"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify == "/spam/egg.pem"
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = EventBridgePutRuleOperator(
            task_id="events_put_rule_job", name=RULE_NAME, event_pattern=EVENT_PATTERN
        )
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(EventBridgeHook, "conn")
    def test_execute(self, mock_conn: MagicMock):
        hook_response = {"RuleArn": "arn:aws:events:us-east-1:123456789012:rule/test"}
        mock_conn.put_rule.return_value = hook_response

        operator = EventBridgePutRuleOperator(
            task_id="events_put_rule_job",
            name=RULE_NAME,
            event_pattern=EVENT_PATTERN,
        )

        result = operator.execute(context={})

        assert result == hook_response

    def test_put_rule_with_bad_json_fails(self):
        operator = EventBridgePutRuleOperator(
            task_id="failed_put_rule_job",
            name=RULE_NAME,
            event_pattern="invalid json",
        )

        with pytest.raises(ValueError):
            operator.execute(None)

    def test_template_fields(self):
        operator = EventBridgePutRuleOperator(
            task_id="events_put_rule_job", name=RULE_NAME, event_pattern=EVENT_PATTERN
        )
        validate_template_fields(operator)


class TestEventBridgeEnableRuleOperator:
    def test_init(self):
        op = EventBridgeEnableRuleOperator(
            task_id="enable_rule_task",
            name=RULE_NAME,
            aws_conn_id="fake-conn-id",
            region_name="us-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.name == RULE_NAME
        assert op.hook.client_type == "events"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "us-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = EventBridgeEnableRuleOperator(task_id="enable_rule_task", name=RULE_NAME)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(EventBridgeHook, "conn")
    def test_enable_rule(self, mock_conn: MagicMock):
        enable_rule = EventBridgeEnableRuleOperator(
            task_id="events_enable_rule_job",
            name=RULE_NAME,
        )

        enable_rule.execute(context={})
        mock_conn.enable_rule.assert_called_with(Name=RULE_NAME)

    def test_template_fields(self):
        operator = EventBridgeEnableRuleOperator(
            task_id="events_enable_rule_job",
            name=RULE_NAME,
        )
        validate_template_fields(operator)


class TestEventBridgeDisableRuleOperator:
    def test_init(self):
        op = EventBridgeDisableRuleOperator(
            task_id="disable_rule_task",
            name=RULE_NAME,
            aws_conn_id="fake-conn-id",
            region_name="ca-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )
        assert op.name == RULE_NAME
        assert op.hook.client_type == "events"
        assert op.hook.resource_type is None
        assert op.hook.aws_conn_id == "fake-conn-id"
        assert op.hook._region_name == "ca-west-1"
        assert op.hook._verify is True
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = EventBridgeDisableRuleOperator(task_id="disable_rule_task", name=RULE_NAME)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(EventBridgeHook, "conn")
    def test_disable_rule(self, mock_conn: MagicMock):
        disable_rule = EventBridgeDisableRuleOperator(
            task_id="events_disable_rule_job",
            name=RULE_NAME,
        )

        disable_rule.execute(context={})
        mock_conn.disable_rule.assert_called_with(Name=RULE_NAME)

    def test_template_fields(self):
        operator = EventBridgeDisableRuleOperator(
            task_id="events_disable_rule_job",
            name=RULE_NAME,
        )
        validate_template_fields(operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.utils.suppress import return_on_error


@pytest.mark.db_test
def test_suppress_function(caplog):
    @return_on_error("error")
    def fn(value: str, exc: Exception | None = None) -> str:
        if exc:
            raise exc
        return value

    caplog.set_level("DEBUG", "airflow.providers.amazon.aws.utils.suppress")
    caplog.clear()

    assert fn("no-error") == "no-error"
    assert not caplog.messages

    assert fn("foo", ValueError("boooo")) == "error"
    assert "Encountered error during execution function/method 'fn'" in caplog.messages

    caplog.clear()
    with pytest.raises(SystemExit, match="42"):
        # We do not plan to catch exception which only based on `BaseExceptions`
        fn("bar", SystemExit(42))
    assert not caplog.messages

    # We catch even serious exceptions, e.g. we do not provide mandatory argument here
    assert fn() == "error"
    assert "Encountered error during execution function/method 'fn'" in caplog.messages


def test_suppress_methods():
    class FakeClass:
        @return_on_error("Oops!… I Did It Again")
        def some_method(self, value, exc: Exception | None = None) -> str:
            if exc:
                raise exc
            return value

        @staticmethod
        @return_on_error(0)
        def some_staticmethod(value, exc: Exception | None = None) -> int:
            if exc:
                raise exc
            return value

        @classmethod
        @return_on_error("It's fine")
        def some_classmethod(cls, value, exc: Exception | None = None) -> str:
            if exc:
                raise exc
            return value

    assert FakeClass().some_method("no-error") == "no-error"
    assert FakeClass.some_staticmethod(42) == 42
    assert FakeClass.some_classmethod("really-no-error-here") == "really-no-error-here"

    assert FakeClass().some_method("foo", KeyError("foo")) == "Oops!… I Did It Again"
    assert FakeClass.some_staticmethod(42, RuntimeError("bar")) == 0
    assert FakeClass.some_classmethod("bar", OSError("Windows detected!")) == "It's fine"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import zipfile
from datetime import datetime
from io import BytesIO

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.lambda_function import (
    LambdaCreateFunctionOperator,
    LambdaInvokeFunctionOperator,
)
from airflow.providers.amazon.aws.sensors.lambda_function import LambdaFunctionStateSensor
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder, prune_logs

DAG_ID = "example_lambda"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

CODE_CONTENT = """
def test(*args):
    print('Hello')
"""


# Create a zip file containing one file "lambda_function.py" to deploy to the lambda function
def create_zip(content: str):
    with BytesIO() as zip_output:
        with zipfile.ZipFile(zip_output, "w", zipfile.ZIP_DEFLATED) as zip_file:
            info = zipfile.ZipInfo("lambda_function.py")
            info.external_attr = 0o777 << 16
            zip_file.writestr(info, content)
        zip_output.seek(0)
        return zip_output.read()


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_lambda(function_name: str):
    client = boto3.client("lambda")
    client.delete_function(
        FunctionName=function_name,
    )


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()

    lambda_function_name: str = f"{test_context[ENV_ID_KEY]}-function"
    role_arn = test_context[ROLE_ARN_KEY]

    # [START howto_operator_create_lambda_function]
    create_lambda_function = LambdaCreateFunctionOperator(
        task_id="create_lambda_function",
        function_name=lambda_function_name,
        runtime="python3.9",
        role=role_arn,
        handler="lambda_function.test",
        code={
            "ZipFile": create_zip(CODE_CONTENT),
        },
    )
    # [END howto_operator_create_lambda_function]

    # [START howto_sensor_lambda_function_state]
    wait_lambda_function_state = LambdaFunctionStateSensor(
        task_id="wait_lambda_function_state",
        function_name=lambda_function_name,
    )
    # [END howto_sensor_lambda_function_state]
    wait_lambda_function_state.poke_interval = 1

    # [START howto_operator_invoke_lambda_function]
    invoke_lambda_function = LambdaInvokeFunctionOperator(
        task_id="invoke_lambda_function",
        function_name=lambda_function_name,
        payload=json.dumps({"SampleEvent": {"SampleData": {"Name": "XYZ", "DoB": "1993-01-01"}}}),
    )
    # [END howto_operator_invoke_lambda_function]

    log_cleanup = prune_logs(
        [
            # Format: ('log group name', 'log stream prefix')
            (f"/aws/lambda/{lambda_function_name}", None),
        ],
        force_delete=True,
        retry=True,
    )

    chain(
        # TEST SETUP
        test_context,
        # TEST BODY
        create_lambda_function,
        wait_lambda_function_state,
        invoke_lambda_function,
        # TEST TEARDOWN
        delete_lambda(lambda_function_name),
        log_cleanup,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.utils.helpers import prune_dict


def _validate_json(pattern: str) -> None:
    try:
        json.loads(pattern)
    except ValueError:
        raise ValueError("`event_pattern` must be a valid JSON string.")


class EventBridgeHook(AwsBaseHook):
    """Amazon EventBridge Hook."""

    def __init__(self, *args, **kwargs):
        super().__init__(client_type="events", *args, **kwargs)

    def put_rule(
        self,
        name: str,
        description: str | None = None,
        event_bus_name: str | None = None,
        event_pattern: str | None = None,
        role_arn: str | None = None,
        schedule_expression: str | None = None,
        state: str | None = None,
        tags: list[dict] | None = None,
        **kwargs,
    ):
        """
        Create or update an EventBridge rule.

        :param name: name of the rule to create or update (required)
        :param description: description of the rule
        :param event_bus_name: name or ARN of the event bus to associate with this rule
        :param event_pattern: pattern of events to be matched to this rule
        :param role_arn: the Amazon Resource Name of the IAM role associated with the rule
        :param schedule_expression: the scheduling expression (for example, a cron or rate expression)
        :param state: indicates whether rule is set to be "ENABLED" or "DISABLED"
        :param tags: list of key-value pairs to associate with the rule

        """
        if not (event_pattern or schedule_expression):
            raise ValueError(
                "One of `event_pattern` or `schedule_expression` are required in order to "
                "put or update your rule."
            )

        if state and state not in ["ENABLED", "DISABLED"]:
            raise ValueError("`state` must be specified as ENABLED or DISABLED.")

        if event_pattern:
            _validate_json(event_pattern)

        put_rule_kwargs: dict[str, str | list] = {
            **prune_dict(
                {
                    "Name": name,
                    "Description": description,
                    "EventBusName": event_bus_name,
                    "EventPattern": event_pattern,
                    "RoleArn": role_arn,
                    "ScheduleExpression": schedule_expression,
                    "State": state,
                    "Tags": tags,
                }
            )
        }

        return self.conn.put_rule(**put_rule_kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest

from airflow.providers.amazon.aws.hooks.kinesis_analytics import KinesisAnalyticsV2Hook
from airflow.providers.amazon.aws.triggers.kinesis_analytics import (
    KinesisAnalyticsV2ApplicationOperationCompleteTrigger,
)
from airflow.triggers.base import TriggerEvent

from unit.amazon.aws.utils.test_waiter import assert_expected_waiter_type

BASE_TRIGGER_CLASSPATH = "airflow.providers.amazon.aws.triggers.kinesis_analytics."


class TestKinesisAnalyticsV2ApplicationOperationCompleteTrigger:
    APPLICATION_NAME = "demo"

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = KinesisAnalyticsV2ApplicationOperationCompleteTrigger(
            application_name=self.APPLICATION_NAME, waiter_name="application_start_complete"
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "KinesisAnalyticsV2ApplicationOperationCompleteTrigger"
        assert kwargs.get("application_name") == self.APPLICATION_NAME

    @pytest.mark.asyncio
    @mock.patch.object(KinesisAnalyticsV2Hook, "get_waiter")
    @mock.patch.object(KinesisAnalyticsV2Hook, "get_async_conn")
    async def test_run_success_with_application_start_complete_waiter(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = KinesisAnalyticsV2ApplicationOperationCompleteTrigger(
            application_name=self.APPLICATION_NAME, waiter_name="application_start_complete"
        )

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent({"status": "success", "application_name": self.APPLICATION_NAME})
        assert_expected_waiter_type(mock_get_waiter, "application_start_complete")
        mock_get_waiter().wait.assert_called_once()

    @pytest.mark.asyncio
    @mock.patch.object(KinesisAnalyticsV2Hook, "get_waiter")
    @mock.patch.object(KinesisAnalyticsV2Hook, "get_async_conn")
    async def test_run_success_with_application_stop_complete_waiter(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = KinesisAnalyticsV2ApplicationOperationCompleteTrigger(
            application_name=self.APPLICATION_NAME, waiter_name="application_stop_waiter"
        )

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent({"status": "success", "application_name": self.APPLICATION_NAME})
        assert_expected_waiter_type(mock_get_waiter, "application_stop_waiter")
        mock_get_waiter().wait.assert_called_once()
